# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    xrange = range
    
"""Unit test for lsl.reader modules that are for ADP stations"""

import os
import unittest

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import tbf
from lsl.reader import cor
from lsl.reader import errors


__revision__ = "$Rev: 2480 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


tbfFile = os.path.join(dataPath, 'tests', 'tbf-test.dat')
corFile = os.path.join(dataPath, 'tests', 'cor-test.dat')


class reader_adp_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.reader
    modules."""
    
    ### TBF ###
    
    def test_tbf_read(self):
        """Test reading in a frame from a TBF file."""
        
        fh = open(tbfFile, 'rb')
        # First frame is really TBF and stores the first channel
        frame1 = tbf.readFrame(fh)
        self.assertTrue(frame1.header.isTBF())
        self.assertEqual(frame1.header.firstChan, 2348)
        # Second frame
        frame2 = tbf.readFrame(fh)
        self.assertTrue(frame2.header.isTBF())
        self.assertEqual(frame2.header.firstChan, 2360)
        fh.close()
        
    def test_tbf_errors(self):
        """Test TBF reading errors."""
        
        fh = open(tbfFile, 'rb')
        # Frames 1 through 5
        for i in range(1,6):
            frame = tbf.readFrame(fh)
            
        # Last frame should be an error (errors.eofError)
        self.assertRaises(errors.eofError, tbf.readFrame, fh)
        fh.close()
        
        # If we offset in the file by 1 byte, we should be a 
        # sync error (errors.syncError).
        fh = open(tbfFile, 'rb')
        fh.seek(1)
        self.assertRaises(errors.syncError, tbf.readFrame, fh)
        fh.close()
        
    def test_tbf_comps(self):
        """Test the TBF frame comparison operators (>, <, etc.) for time tags."""
        
        fh = open(tbfFile, 'rb')
        # Frames 1 through 4
        frames = []
        for i in range(1,5):
            frames.append(tbf.readFrame(fh))
        fh.close()
        
        self.assertTrue(0 < frames[0])
        self.assertFalse(0 > frames[0])
        self.assertTrue(frames[-1] >= frames[0])
        self.assertTrue(frames[-1] <= frames[0])
        self.assertTrue(frames[0] == frames[0])
        self.assertTrue(frames[0] == frames[-1])
        self.assertFalse(frames[0] != frames[0])
        
    def test_tbf_sort(self):
        """Test sorting TBF frames by time tags."""
        
        fh = open(tbfFile, 'rb')
        # Frames 1 through 3
        frames = []
        for i in range(1,4):
            frames.append(tbf.readFrame(fh))
        fh.close()
        
        frames.sort()
        frames = frames[::-1]
        
        for i in xrange(1,len(frames)):
            self.assertTrue( frames[i-1] >= frames[i] )
            
    def test_tbf_math(self):
        """Test mathematical operations on TBF frame data via frames."""
        
        fh = open(tbfFile, 'rb')
        # Frames 1 through 3
        frames = []
        for i in range(1,4):
            frames.append(tbf.readFrame(fh))
        fh.close()
        
        # Multiplication
        frameT = frames[0] * 2.0
        for i in range(12*256*2):
            c = i // 2 // 256
            s = i // 2 % 256
            p = i % 2
            self.assertAlmostEqual(frameT.data.fDomain[c,s,p], 2*frames[0].data.fDomain[c,s,p], 2)
        frameT *= 2.0
        for i in range(12*256*2):
            c = i // 2 // 256
            s = i // 2 % 256
            p = i % 2
            self.assertAlmostEqual(frameT.data.fDomain[c,s,p], 4*frames[0].data.fDomain[c,s,p], 2)
        frameT = frames[0] * frames[1]
        for i in range(12*256*2):
            c = i // 2 // 256
            s = i // 2 % 256
            p = i % 2
            self.assertAlmostEqual(frameT.data.fDomain[c,s,p], frames[0].data.fDomain[c,s,p]*frames[1].data.fDomain[c,s,p], 2)
            
        # Addition
        frameA = frames[0] + 2.0
        for i in range(800):
            c = i // 2 // 256
            s = i // 2 % 256
            p = i % 2
            self.assertAlmostEqual(frameA.data.fDomain[c,s,p], 2+frames[0].data.fDomain[c,s,p], 2)
        frameA += 2.0
        for i in range(800):
            c = i // 2 // 256
            s = i // 2 % 256
            p = i % 2
            self.assertAlmostEqual(frameA.data.fDomain[c,s,p], 4+frames[0].data.fDomain[c,s,p], 2)
        frameA = frames[0] + frames[1]
        for i in range(800):
            c = i // 2 // 256
            s = i // 2 % 256
            p = i % 2
            self.assertAlmostEqual(frameA.data.fDomain[c,s,p], frames[0].data.fDomain[c,s,p]+frames[1].data.fDomain[c,s,p], 2)
            
     ### COR ###
    
    def test_cor_read(self):
        """Test reading in a frame from a COR file."""
        
        fh = open(corFile, 'rb')
        # First frame is really COR and check the basic metadata
        frame1 = cor.readFrame(fh)
        self.assertTrue(frame1.header.isCOR())
        self.assertEqual(frame1.header.firstChan, 1584)
        self.assertEqual(frame1.parseID(), (1,1))
        self.assertEqual(frame1.getIntegrationTime(), 5)
        self.assertEqual(frame1.getGain(), 1)
        # Second frame
        frame2 = cor.readFrame(fh)
        self.assertTrue(frame2.header.isCOR())
        self.assertEqual(frame2.header.firstChan, 1584)
        self.assertEqual(frame2.parseID(), (1,2))
        self.assertEqual(frame2.getIntegrationTime(), 5)
        self.assertEqual(frame2.getGain(), 1)
        fh.close()
        
    def test_cor_errors(self):
        """Test COR reading errors."""
        
        fh = open(corFile, 'rb')
        # Frames 1 through 65
        for i in range(1,66):
            frame = cor.readFrame(fh)
            
        # Last frame should be an error (errors.eofError)
        self.assertRaises(errors.eofError, cor.readFrame, fh)
        fh.close()
        
        # If we offset in the file by 1 byte, we should be a 
        # sync error (errors.syncError).
        fh = open(corFile, 'rb')
        fh.seek(1)
        self.assertRaises(errors.syncError, cor.readFrame, fh)
        fh.close()
        
    def test_cor_frames(self):
        """Test determing the number of frames per observation in a COR file."""
        
        fh = open(corFile, 'rb')
        self.assertEqual(32896, cor.getFramesPerObs(fh))
        fh.close()
        
    def test_cor_channels(self):
        """Test determing the number of channels in a COR observation."""
        
        fh = open(corFile, 'rb')
        self.assertEqual(72, cor.getChannelCount(fh))
        fh.close()
        
    def test_cor_baselines(self):
        """Test determing the number of baselines in a COR observation."""
        
        fh = open(corFile, 'rb')
        self.assertEqual(32896, cor.getBaselineCount(fh))
        fh.close()
        
    def test_cor_comps(self):
        """Test the COR frame comparison operators (>, <, etc.) for time tags."""
        
        fh = open(corFile, 'rb')
        # Frames 1 through 29
        frames = []
        for i in range(1,30):
            frames.append(cor.readFrame(fh))
        fh.close()
        
        self.assertTrue(0 < frames[0])
        self.assertFalse(0 > frames[0])
        self.assertTrue(frames[-1] >= frames[0])
        self.assertTrue(frames[-1] <= frames[0])
        self.assertTrue(frames[0] == frames[0])
        self.assertFalse(frames[0] != frames[-1])
        self.assertFalse(frames[0] != frames[0])
        
    def test_cor_sort(self):
        """Test sorting COR frames by time tags."""
        
        fh = open(corFile, 'rb')
        # Frames 1 through 29
        frames = []
        for i in range(1,30):
            frames.append(cor.readFrame(fh))
        fh.close()
        
        frames.sort()
        frames = frames[::-1]
        
        for i in xrange(1,len(frames)):
            self.assertTrue( frames[i-1] >= frames[i] )
            
    def test_cor_math(self):
        """Test mathematical operations on COR frame data via frames."""
        
        fh = open(corFile, 'rb')
        # Frames 1 through 29
        frames = []
        for i in range(1,30):
            frames.append(cor.readFrame(fh))
        fh.close()
        
        # Multiplication
        frameT = frames[0] * 2.0
        for i in range(72):
            for j in xrange(2):
                for k in xrange(2):
                    self.assertAlmostEqual(frameT.data.vis[i,j,k], 2*frames[0].data.vis[i,j,k], 2)
        frameT *= 2.0
        for i in range(72):
            for j in xrange(2):
                for k in xrange(2):
                    self.assertAlmostEqual(frameT.data.vis[i,j,k], 4*frames[0].data.vis[i,j,k], 2)
        frameT = frames[0] * frames[1]
        for i in range(72):
            for j in xrange(2):
                for k in xrange(2):
                    self.assertAlmostEqual(frameT.data.vis[i,j,k], frames[0].data.vis[i,j,k]*frames[1].data.vis[i,j,k], 2)
            
        # Addition
        frameA = frames[0] + 2.0
        for i in range(72):
            for j in xrange(2):
                for k in xrange(2):
                    self.assertAlmostEqual(frameA.data.vis[i,j,k], 2+frames[0].data.vis[i,j,k], 2)
        frameA += 2.0
        for i in range(72):
            for j in xrange(2):
                for k in xrange(2):
                    self.assertAlmostEqual(frameA.data.vis[i,j,k], 4+frames[0].data.vis[i,j,k], 2)
        frameA = frames[0] + frames[1]
        for i in range(72):
            for j in xrange(2):
                for k in xrange(2):
                    self.assertAlmostEqual(frameA.data.vis[i,j,k], frames[0].data.vis[i,j,k]+frames[1].data.vis[i,j,k], 2)
            
        

class reader_adp_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.reader units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(reader_adp_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for lsl.misc.parser module."""


import unittest
import ephem
import numpy
import argparse

from lsl.misc import parser
from lsl.common.constants import c


__revision__  = "$Rev: 2640 $"
__version__   = "0.1"
__author__    = "Jayce Dowell"


class parser_tests(unittest.TestCase):
    def test_positive_or_zero_int(self):
        """Test the parser.positive_or_zero_int conversion function."""
        
        self.assertEqual(parser.positive_or_zero_int('5'), 5)
        self.assertEqual(parser.positive_or_zero_int('0'), 0)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_or_zero_int, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_or_zero_int, '1.5')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_or_zero_int, '-1')
        
    def test_positive_int(self):
        """Test the parser.positive_int conversion function."""
        
        self.assertEqual(parser.positive_int('5'), 5)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_int, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_int, '1.5')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_int, '0')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_int, '-1')
        
    def test_positive_or_zero_float(self):
        """Test the parser.positive_or_zero_float conversion function."""
        
        self.assertEqual(parser.positive_or_zero_float('5'), 5.0)
        self.assertEqual(parser.positive_or_zero_float('5.1'), 5.1)
        self.assertEqual(parser.positive_or_zero_float('0'), 0.0)
        self.assertEqual(parser.positive_or_zero_float('0.0'), 0.0)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_or_zero_float, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_or_zero_float, '-1.1')
        
    def test_positive_float(self):
        """Test the parser.positive_float conversion function."""
        
        self.assertEqual(parser.positive_float('5'), 5.0)
        self.assertEqual(parser.positive_float('5.1'), 5.1)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_float, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_float, '0')
        self.assertRaises(argparse.ArgumentTypeError, parser.positive_float, '-1.1')
        
    def test_frequency(self):
        """Test the parser.frequency conversion function."""
        
        self.assertEqual(parser.frequency('45'), 45e6)
        self.assertEqual(parser.frequency('45kHz'), 45e3)
        self.assertEqual(parser.frequency('45.6 GHz'), 45.6e9)
        self.assertEqual(parser.frequency('4.4m'), c/4.4)
        self.assertEqual(parser.frequency('21cm'), c/0.21)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.frequency, '45G')
        self.assertRaises(argparse.ArgumentTypeError, parser.frequency, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.frequency, '4~5')
        
    def test_frequency_range(self):
        """Test the parser.frequency_range conversion function."""
        
        self.assertEqual(parser.frequency_range('4~5')[0], 4e6)
        self.assertEqual(parser.frequency_range('4~5')[1], 5e6)
        self.assertEqual(parser.frequency_range('4~5GHz')[0], 4e9)
        self.assertEqual(parser.frequency_range('4~5GHz')[1], 5e9)
        self.assertEqual(parser.frequency_range('4MHz~5GHz')[0], 4e6)
        self.assertEqual(parser.frequency_range('5GHz~4MHz')[1], 5e9)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.frequency_range, '45MHz~50')
        self.assertRaises(argparse.ArgumentTypeError, parser.frequency_range, '4~c')
        self.assertRaises(argparse.ArgumentTypeError, parser.frequency_range, '4')
        
    def test_wavelength(self):
        """Test the parser.wavelength conversion function."""
        
        self.assertEqual(parser.wavelength('45'), 45.0)
        self.assertEqual(parser.wavelength('45km'), 45e3)
        self.assertEqual(parser.wavelength('45.6 Gm'), 45.6e9)
        self.assertEqual(parser.wavelength('47MHz'), c/47e6)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.wavelength, '45G')
        self.assertRaises(argparse.ArgumentTypeError, parser.wavelength, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.wavelength, '4~5')
        
    def test_wavelength_range(self):
        """Test the parser.wavelength_range conversion function."""
        
        self.assertEqual(parser.wavelength_range('4~5')[0], 4.0)
        self.assertEqual(parser.wavelength_range('4~5')[1], 5.0)
        self.assertEqual(parser.wavelength_range('4~5cm')[0], 4e-2)
        self.assertEqual(parser.wavelength_range('4~5cm')[1], 5e-2)
        self.assertEqual(parser.wavelength_range('4cm~5m')[0], 4e-2)
        self.assertEqual(parser.wavelength_range('5m~4cm')[1], 5.0)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.wavelength_range, '45Mm~50')
        self.assertRaises(argparse.ArgumentTypeError, parser.wavelength_range, '4~c')
        self.assertRaises(argparse.ArgumentTypeError, parser.wavelength_range, '4')
        
    def test_date(self):
        """Test the parser.date conversion function."""
        
        self.assertEqual(parser.date('2001/1/1'),   '2001/01/01')
        self.assertEqual(parser.date('2010/02/23'), '2010/02/23')
        self.assertEqual(parser.date('2008/02/29'), '2008/02/29')
        self.assertEqual(parser.date('2001/1/1'),   '2001/01/01')
        self.assertEqual(parser.date('58427'),      '2018/11/05')
        
        self.assertRaises(argparse.ArgumentTypeError, parser.date, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.date, '2018/1/32')
        self.assertRaises(argparse.ArgumentTypeError, parser.date, '2018/13/1')
        self.assertRaises(argparse.ArgumentTypeError, parser.date, '2018/2/29')
        
    def test_mjd(self):
        """Test the parser.mjd conversion function."""
        
        self.assertEqual(parser.mjd('56274'),     56274)
        self.assertEqual(parser.mjd('2018/11/5'), 58427)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.mjd, 'c')
        self.assertRaises(argparse.ArgumentTypeError, parser.mjd, '2018/1/32')
        self.assertRaises(argparse.ArgumentTypeError, parser.mjd, '2018/13/1')
        
    def test_time(self):
        """Test the parser.time conversion function."""
        
        self.assertEqual(parser.time('1:23:45.6'),  '1:23:45.600000')
        self.assertEqual(parser.time('1:23:45'),    '1:23:45.000000')
        self.assertEqual(parser.time('23:59:59'),  '23:59:59.000000')
        self.assertEqual(parser.time('3615100'),    '1:00:15.100000')
        
        self.assertRaises(argparse.ArgumentTypeError, parser.time, '24:58:23.5')
        self.assertRaises(argparse.ArgumentTypeError, parser.time, '-1:58:23.5')
        self.assertRaises(argparse.ArgumentTypeError, parser.time, '-1')
        self.assertRaises(argparse.ArgumentTypeError, parser.time, '86500000')
        
    def test_mpm(self):
        """Test the parser.mpm conversion function."""
        
        self.assertEqual(parser.mpm('1:23:45.6'),       5025600)
        self.assertEqual(parser.mpm('1:23:45'),         5025000)
        self.assertEqual(parser.mpm('23:59:59'),       86399000)
        self.assertEqual(parser.mpm('1:00:15.100000'),  3615100)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.mpm, '24:58:23.5')
        self.assertRaises(argparse.ArgumentTypeError, parser.mpm, '-1:58:23.5')
        self.assertRaises(argparse.ArgumentTypeError, parser.mpm, '-1')
        self.assertRaises(argparse.ArgumentTypeError, parser.mpm, '86500000')
        
    def test_hours(self):
        """Test the parser.hours conversion function."""
        
        self.assertAlmostEqual(parser.hours('5'), 5*numpy.pi/12, 15)
        self.assertAlmostEqual(parser.hours('5:01:02.3'), (5+1/60.+2.3/3600.)*numpy.pi/12, 15)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.hours, 'c')
        
    def test_csv_hours_list(self):
        """Test the parser.hours conversion function."""
        
        eh = ephem.hours
        
        cin, cout = '5,', [eh('5:00:00'),]
        cin = parser.csv_hours_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertAlmostEqual(cin[i], cout[i], 15)
            
        cin, cout = '5,6.5,7:10,8:10:20.1', [eh('5:00:00'),eh('6:30:00'),eh('7:10:00'),eh('8:10:20.1')]
        cin = parser.csv_hours_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertAlmostEqual(cin[i], cout[i], 15)
            
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_hours_list, '5,6,c')
        
    def test_degrees(self):
        """Test the parser.degrees conversion function."""
        
        self.assertAlmostEqual(parser.degrees('5'), 5*numpy.pi/180, 15)
        self.assertAlmostEqual(parser.degrees('5:01:02.3'), (5+1/60.+2.3/3600.)*numpy.pi/180, 15)
        
        self.assertRaises(argparse.ArgumentTypeError, parser.degrees, 'c')
        
    def test_csv_degrees_list(self):
        """Test the parser.degrees conversion function."""
        
        ed = ephem.degrees
        
        cin, cout = '5,', [ed('5:00:00'),]
        cin = parser.csv_degrees_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertAlmostEqual(cin[i], cout[i], 15)
            
        cin, cout = '5,-6.5,7:10,8:10:20.1', [ed('5:00:00'),ed('-6:30:00'),ed('7:10:00'),ed('8:10:20.1')]
        cin = parser.csv_degrees_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertAlmostEqual(cin[i], cout[i], 15)
            
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_degrees_list, '5,6,c')
        
    def test_csv_int_list(self):
        """Test the parser.csv_int_list conversion function."""
        
        cin, cout = '*', 'all'
        self.assertEqual(parser.csv_int_list(cin), cout)
        
        cin, cout = '', 'none'
        self.assertEqual(parser.csv_int_list(cin), cout)
        
        cin, cout = '1', [1,]
        cin = parser.csv_int_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = '1,2,', [1,2]
        cin = parser.csv_int_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = '10~12,1,2,5~8', [10,11,12,1,2,5,6,7,8]
        cin = parser.csv_int_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_int_list, '1,2,!5')
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_int_list, '1,2,c')
        
    def test_csv_baseline_list(self):
        """Test the parser.csv_baseline_list conversion function."""
        
        cin, cout = '*', 'all'
        self.assertEqual(parser.csv_baseline_list(cin), cout)
        
        cin, cout = '', 'none'
        self.assertEqual(parser.csv_baseline_list(cin), cout)
        
        cin, cout = '1-2,', [(1,2),]
        cin = parser.csv_baseline_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            for j in range(2):
                self.assertEqual(cin[i][j], cout[i][j])
                
        cin, cout = '10-12,1-2,1-3~4,2~3-3~4', [(10,12),(1,2),(1,3),(1,4),(2,3),(2,4),(3,3),(3,4)]
        cin = parser.csv_baseline_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            for j in range(2):
                self.assertEqual(cin[i][j], cout[i][j])
            
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_baseline_list, '1')
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_baseline_list, '1-2,2')
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_baseline_list, '1-2,1-!2')
        
    def test_csv_hostname_list(self):
        """Test the parser.csv_hostname_list conversion function."""
        
        cin, cout = 'host1', ['host1',]
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = 'host-1,host-2', ['host-1','host-2']
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = 'host1,host2,host5~6', ['host1','host2','host5','host6']
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_hostname_list, 'c&')
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_hostname_list, '(9')
        
        cin, cout = '192.168.1.1', ['192.168.1.1',]
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = '192.168.1.1,192.168.1.2,192.168.1.5~6', ['192.168.1.1','192.168.1.2','192.168.1.5','192.168.1.6']
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = '192.168.5~6.1', ['192.168.5.1','192.168.6.1']
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = '192.168~169.1.1~2,', ['192.168.1.1','192.168.1.2','192.169.1.1','192.169.1.2']
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        cin, cout = '192~193.168.5~6.1', ['192.168.5.1','192.168.6.1','193.168.5.1','193.168.6.1']
        cin = parser.csv_hostname_list(cin)
        self.assertEqual(len(cin), len(cout))
        for i in range(len(cout)):
            self.assertEqual(cin[i], cout[i])
            
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_baseline_list, '192.168.1')
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_baseline_list, '192.168.1.c')
        self.assertRaises(argparse.ArgumentTypeError, parser.csv_baseline_list, '192.168.1.1.1')


class parser_test_suite(unittest.TestSuite):
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(parser_tests))
                

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for the lsl.statistics.robust module."""

import os
import time
import warnings
import unittest
import numpy

from lsl.statistics import robust


__revision__ = "$Rev: 2394 $"
__version__  = "0.2"
__author__    = "Jayce Dowell"

class robust_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.statistics.robust
    module."""
    
    def setUp(self):
        """Turn off all numpy and python warnings."""

        numpy.seterr(all='ignore')
        warnings.simplefilter('ignore')
        
    def test_biweight(self):
        """Test the biweight mean function."""
        
        # Make sure that it can do simple averages
        a = numpy.random.rand(512)
        self.assertAlmostEqual(a.mean(), robust.biweightMean(a), 1)

        b = numpy.random.randn(512)
        self.assertAlmostEqual(b.mean(), robust.biweightMean(b), 1)

        # Make sure it can reject obvious points
        b = 1.0*a
        b[10] = 1e6
        self.assertTrue(robust.biweightMean(b) < b.mean())

    def test_mean(self):
        """Test the outlier-resistant mean function."""

        # Make sure that it can do simple averages
        a = numpy.random.rand(512)
        self.assertAlmostEqual(a.mean(), robust.mean(a, Cut=100), 6)

        b = numpy.random.randn(512)
        self.assertAlmostEqual(b.mean(), robust.mean(b, Cut=100), 6)

        # Make sure it can reject obvious points
        b = 1.0*a
        b[10] = 1e6
        self.assertTrue(robust.mean(b) < b.mean())

    def test_std(self):
        """Test the outlier-resistant standard deviation function."""

        # Make sure it can reject obvious points
        b = numpy.random.randn(512)
        b[10] = 1e6
        self.assertTrue(robust.std(b) < b.std())

    def test_linefit(self):
        """Test the outlier-resistant line fitter."""

        x = numpy.linspace(-10, 10, 256)
        y = 1.34*x + 0.56

        cc = robust.linefit(x, y)
        self.assertAlmostEqual(cc[0], 1.34, 2)
        self.assertAlmostEqual(cc[1], 0.56, 2)

        cc = robust.linefit(x, y, Bisector=True)
        self.assertAlmostEqual(cc[0], 1.34, 2)
        self.assertAlmostEqual(cc[1], 0.56, 2)

        x = numpy.linspace(-10, 10, 2048)
        y = 2.86*x - 0.56

        cc = robust.linefit(x, y)
        self.assertAlmostEqual(cc[0], 2.86,  2)
        self.assertAlmostEqual(cc[1], -0.56, 2)

        cc = robust.linefit(x, y, Bisector=True)
        self.assertAlmostEqual(cc[0], 2.86,  2)
        self.assertAlmostEqual(cc[1], -0.56, 2)

    def test_polyfit(self):
        """Test the outlier-resistant polynomial fitter."""

        x = numpy.linspace(-10, 10, 256)
        y = 0.012*x**3 + 0.34*x**2 + 1.34*x + 0.56

        cc = robust.polyfit(x, y, 3)
        self.assertAlmostEqual(cc[0], 0.012, 3)
        self.assertAlmostEqual(cc[1], 0.340, 3)
        self.assertAlmostEqual(cc[2], 1.340, 3)
        self.assertAlmostEqual(cc[3], 0.560, 3)

        x = numpy.linspace(-10, 10, 2048)
        y = 0.003*x**4 + 0.012*x**3 + 0.34*x**2 + 1.34*x + 0.56

        cc = robust.polyfit(x, y, 4)
        self.assertAlmostEqual(cc[0], 0.003, 3)
        self.assertAlmostEqual(cc[1], 0.012, 3)
        self.assertAlmostEqual(cc[2], 0.340, 3)
        self.assertAlmostEqual(cc[3], 1.340, 3)
        self.assertAlmostEqual(cc[4], 0.560, 3)


class robust_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.statistics.robust 
    units tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(robust_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    long = int

"""
Module to simulate observations made with the DP system.
"""

import time
import numpy
from aipy import coord as aipycoord

from lsl import astro
from lsl.common import dp as dp_common
from lsl.common import stations as lwa_common
from lsl.sim import tbw
from lsl.sim import tbn
from lsl.sim import drx
from lsl.sim import vis
from lsl.reader.tbn import filterCodes as TBNFilters
from lsl.reader.drx import filterCodes as DRXFilters

__version__ = '0.3'
__revision__ = '$Rev: 2467 $'
__all__ = ['basicSignal', 'pointSource', '__version__', '__revision__', '__all__']


def __basicTBW(fh, stands, nFrames, **kwargs):
    """
    Private function for generating a basic TBW signal.
    """

    tStart = kwargs['tStart']
    bits = kwargs['bits']
    verbose = kwargs['verbose']
    noiseStrength = kwargs['noiseStrength']
    
    if bits == 12:
        maxValue = 2047
        samplesPerFrame = 400
    else:
        maxValue =  7
        samplesPerFrame = 1200

    if verbose:
        print("Simulating %i captures of %i-bit TBW data for %i stands:" % \
            (int(numpy.ceil(nFrames / 30000.0)), bits, len(stands)))

    nCaptures = int(numpy.ceil(nFrames / 30000.0))
    for capture in xrange(nCaptures):
        for stand1, stand2 in zip(stands[0::2], stands[1::2]):
            FramesThisBatch = nFrames - capture*30000
            if FramesThisBatch > 30000:
                FramesThisBatch = 30000
            if verbose:
                print(" capture %i, stands %i and %i" % (capture+1, stand1, stand2))

            for i in xrange(FramesThisBatch):
                t = long(tStart*dp_common.fS) + i*samplesPerFrame
                t += long(60*dp_common.fS*capture)
                tFrame = t/dp_common.fS - tStart + numpy.arange(samplesPerFrame, dtype=numpy.float32) / dp_common.fS
                
                cFrame = tbw.SimFrame(stand=stand1, frameCount=i+1, dataBits=bits, obsTime=t)
                cFrame.xy = numpy.random.randn(2, samplesPerFrame)
                cFrame.xy[0,:] *= maxValue*noiseStrength
                cFrame.xy[0,:] += maxValue*numpy.cos(2*numpy.pi*40.0e6*tFrame)
                cFrame.xy[1,:] *= maxValue*noiseStrength
                cFrame.xy[1,:] += maxValue*numpy.cos(2*numpy.pi*60.0e6*tFrame)
                cFrame.writeRawFrame(fh)

                cFrame = tbw.SimFrame(stand=stand2, frameCount=i+1, dataBits=bits, obsTime=t)
                cFrame.xy = numpy.random.randn(2, samplesPerFrame)
                cFrame.xy[0,:] *= maxValue*noiseStrength
                cFrame.xy[0,:] += maxValue*numpy.cos(2*numpy.pi*30.0e6*tFrame)
                cFrame.xy[1,:] *= maxValue*noiseStrength
                cFrame.xy[1,:] += maxValue*numpy.cos(2*numpy.pi*50.0e6*tFrame)
                cFrame.writeRawFrame(fh)


def __basicTBN(fh, stands, nFrames, **kwargs):
    """
    Private function for generating a basic TBN signal.
    """

    tStart = kwargs['tStart']
    filter = kwargs['filter']
    verbose = kwargs['verbose']
    noiseStrength = kwargs['noiseStrength']
    sampleRate = TBNFilters[filter]
    
    maxValue = 127
    samplesPerFrame = 512
    upperSpike = sampleRate / 4.0
    lowerSpike = -sampleRate / 4.0
    
    if verbose:
        print("Simulating %i frames of TBN Data @ %.2f kHz for %i stands:" % \
            (nFrames, sampleRate/1e3, len(stands)))
    
    for i in xrange(nFrames):
        if i % 1000 == 0 and verbose:
            print(" frame %i" % (i+1))
        t = long(tStart*dp_common.fS) + long(i*dp_common.fS*samplesPerFrame/sampleRate)
        tFrame = t/dp_common.fS - tStart + numpy.arange(samplesPerFrame, dtype=numpy.float32) / sampleRate
        for stand in stands:
            cFrame = tbn.SimFrame(stand=stand, pol=0, centralFreq=40e6, gain=20, frameCount=i+1, obsTime=t)
            cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
            cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
            cFrame.iq *= maxValue*noiseStrength
            cFrame.iq += maxValue*numpy.exp(2j*numpy.pi*upperSpike*tFrame)
            cFrame.writeRawFrame(fh)

            cFrame = tbn.SimFrame(stand=stand, pol=1, centralFreq=40e6, gain=20, frameCount=i+1, obsTime=t)
            cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
            cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
            cFrame.iq *= maxValue*noiseStrength
            cFrame.iq += maxValue*numpy.exp(2j*numpy.pi*lowerSpike*tFrame)
            cFrame.writeRawFrame(fh)

def __basicDRX(fh, stands, nFrames, **kwargs):
    """
    Private function for generating a basic TBN signal.
    """

    tStart = kwargs['tStart']
    filter = kwargs['filter']
    nTuning = kwargs['nTuning']
    verbose = kwargs['verbose']
    noiseStrength = kwargs['noiseStrength']
    sampleRate = DRXFilters[filter]
    
    maxValue = 7
    samplesPerFrame = 4096
    upperSpike1 = sampleRate / 4.0
    lowerSpike1 = -sampleRate / 4.0
    upperSpike2 = sampleRate / 3.0
    lowerSpike2 = -sampleRate / 3.0

    if verbose:
        print("Simulating %i frames of DRX Data @ %.2f MHz for %i beams, %i tunings each:" % \
            (nFrames, sampleRate/1e6, len(stands), nTuning))

    beams = stands
    for i in range(nFrames):
        if i % 1000 == 0 and verbose:
            print(" frame %i" % i)
        t = long(tStart*dp_common.fS) + long(i*dp_common.fS*samplesPerFrame/sampleRate)
        tFrame = t/dp_common.fS - tStart + numpy.arange(samplesPerFrame, dtype=numpy.float32) / sampleRate
        for beam in beams:
            for tune in range(1, nTuning+1):
                if tune == 1:
                    # Tuning 1:
                    cFrame = drx.SimFrame(beam=beam, tune=1, pol=0, frameCount=i+1, filterCode=filter, timeOffset=0, obsTime=t, flags=0)
                    cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
                    cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
                    cFrame.iq *= maxValue*noiseStrength
                    cFrame.iq += maxValue*numpy.exp(2j*numpy.pi*upperSpike1*tFrame)
                    cFrame.writeRawFrame(fh)
            
                    cFrame = drx.SimFrame(beam=beam, tune=1, pol=1, frameCount=i+1, filterCode=filter, timeOffset=0, obsTime=t, flags=0)
                    cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
                    cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
                    cFrame.iq *= maxValue*noiseStrength
                    cFrame.iq += maxValue*numpy.exp(2j*numpy.pi*lowerSpike1*tFrame)
                    cFrame.writeRawFrame(fh)
                else:
                    # Tuning 2:
                    cFrame = drx.SimFrame(beam=beam, tune=2, pol=0, frameCount=i+1, filterCode=filter, timeOffset=0, obsTime=t, flags=0)
                    cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
                    cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
                    cFrame.iq *= maxValue*noiseStrength
                    cFrame.iq += maxValue*numpy.exp(2j*numpy.pi*lowerSpike2*tFrame)
                    cFrame.writeRawFrame(fh)
            
                    cFrame = drx.SimFrame(beam=beam, tune=2, pol=1, frameCount=i+1, filterCode=filter, timeOffset=0, obsTime=t, flags=0)
                    cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
                    cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
                    cFrame.iq *= maxValue*noiseStrength
                    cFrame.iq += maxValue*numpy.exp(2j*numpy.pi*upperSpike2*tFrame)
                    cFrame.writeRawFrame(fh)


def basicSignal(fh, stands, nFrames, mode='DRX', filter=6, nTuning=2, bits=12, tStart=0, noiseStrength=0.1, verbose=False):
    """
    Generate a collection of frames with a basic test signal for TBW, TBN, 
    and DRX.  The signals for the three modes are:
    
    TBW
    * noise + 40 MHz signal for x-pol.
    * noise + 60 MHz signal for y-pol.
        -> odd stands
    * noise + 30 MHz signal for x-pol.
    * noise + 50 MHz signal for ypol.
        -> even stands
    
    TBN
    * noise + (sampleRate/4) kHz signal for x-pol. and noise + 
        (-sampleRate/4) for y-pol.

    DRX
    * noise + (sampleRate/4) kHz signal for x-pol. and noise + 
        (-sampleRate/4) for y-pol. -> tuning 1
    * noise + (-sampleRate/3) kHz signal for x-pol. and noise + 
        (sampleRate/3) for y-pol. -> tuning 2
        
    All modes need to have stands (beams in the case of DRX) and number of
    frames to generate.  TBW also needs to 'bits' keyword set to generate 
    either 12-bit or 4-bit data.  The TBN and DRX frames need the 'filter'
    keyword set to specify the filter width.  In addition, the 'stands' 
    argument is interpreted as beam numbers for DRX.
    
    .. versionchanged:: 0.4.4
        Added the `noiseStrength` keyword to control how much noise is added to 
        the data.
    """

    if tStart == 0:
        tStart = time.time()

    if mode == 'TBW':
        __basicTBW(fh, stands, nFrames, bits=bits, tStart=tStart, noiseStrength=noiseStrength, verbose=verbose)
    elif mode == 'TBN':
        __basicTBN(fh, stands, nFrames, filter=filter, tStart=tStart, noiseStrength=noiseStrength, verbose=verbose)
    elif mode == 'DRX':
        __basicDRX(fh, stands, nFrames, filter=filter, nTuning=nTuning, tStart=tStart, noiseStrength=noiseStrength, verbose=verbose)
    else:
        raise RuntimeError("Unknown observations mode: %s" % mode)


def __getAntennaArray(station, stands, utime, freqs):
    """
    Given a LWA station object, a list of stands, an observation time, and
    a list of frequencies in Hz, build an aipy AntennaArray object.
    """

    return vis.buildSimArray(station, stands, freqs/1e9, jd=astro.unix_to_utcjd(utime))


def __getSourceParameters(aa, timestamp, srcs):
    """
    Given an aipy AntennaArray object, an observation time, and aipy.src 
    object, return all of the parameters needed for a simulation.
    """
    
    # Set the time for the array
    aa.set_unixtime(timestamp)

    # Compute the source parameters
    srcs_tp = []
    srcs_mt = []
    srcs_jy = []
    srcs_fq = []
    for name in srcs:
        ## Update the source's coordinates
        src = srcs[name]
        src.compute(aa)

        ## Get parameters
        top = src.get_crds(crdsys='top', ncrd=3)	# topo. coords.
        mat = src.map							# equitorial -> topo. rotation matrix
        jys = src.get_jys()						# F_nu
        frq = aa.get_afreqs()					# nu

        ## Fix the lowest frequencies to avoid problems with the flux blowing up
        ## at nu = 0 Hz by replacing flux values below 1 MHz with the flux at 
        ## 1 MHz
        Jyat1MHz = jys[ numpy.where( numpy.abs(frq-0.001) == numpy.abs(frq-0.001).min() ) ]
        jys = numpy.where( frq >= 0.001, jys, Jyat1MHz )

        ## Filter out sources that are below the horizon or have no flux
        srcAzAlt = aipycoord.top2azalt(top)
        if srcAzAlt[1] <= 0 or jys.sum() <= 0:
            continue

        ## Save values into the source arrays
        srcs_tp.append( top )
        srcs_mt.append( mat )
        srcs_jy.append( jys )
        srcs_fq.append( frq )

    # Return the values as a dictionary
    return {'topo': srcs_tp, 'trans': srcs_mt, 'flux': srcs_jy, 'freq': srcs_fq}


def __buildSignals(aa, stands, srcParams, times, pol='x', phaseCenter='z'):
    """
    Given an aipy AntennaArray, a list of stand numbers, a dictionary of source 
    parameters, and an array of times in ns, return a numpy array of the simulated 
    signals that is Nstands x Ntimes in shape
    ."""

    # Find out how many stands, srcs, and samples (times) we are working with
    Nstand = len(aa.ants)
    Nsrc = len(srcParams['topo'])
    Ntime = len(times)
    
    # Get the topocentric coorindates for the zenith
    zen = aipycoord.azalt2top(numpy.array([[numpy.pi/4], [numpy.pi/2]]))
    zen = numpy.squeeze(zen)
    
    # Update the phase center if necessary
    if phaseCenter == 'z':
        phaseCenterMap = aipycoord.eq2top_m(0.0, aa.lat)
    else:
        phaseCenter.compute(aa)
        phaseCenterMap = phaseCenter.map

    # Setup a temporary array to hold the signals per source, stand, and time.  
    # This array is complex so that it can accomidate both TBW and TBN data at
    # the same time
    temp = numpy.zeros((Nsrc, Nstand, Ntime), dtype=numpy.complex64)

    # Loop over sources and stands to build up the signals
    srcCount = 0
    for topo,trans,flux,freq in zip(srcParams['topo'], srcParams['trans'], srcParams['flux'], srcParams['freq']):
        antCount = 0
        for ant,std in zip(aa.ants, stands):
            # Zeroth, get the beam response in the direction of the current source for all frequencies
            antResponse = numpy.squeeze( ant.bm_response(topo, pol=pol) )

            # First, do the geometric delay
            geoDelay = ( numpy.dot(trans, ant.pos).transpose() )[2] 
            geoDelayPC = ( numpy.dot(phaseCenterMap, ant.pos).transpose() )[2]

            # Second, do the cable delay
            Delayat1MHz = std.cable.delay(frequency=1.0e6) * 1e9 # s -> ns
            cblDelay = std.cable.delay(frequency=aa.get_afreqs()*1e9) * 1e9 # s -> ns
            # NB: Replace the cable delays below 1 MHz with the 1 MHz value to keep the 
            # delays from blowing up for small f
            cblDelay = numpy.where( freq >= 0.001, cblDelay, Delayat1MHz )

            for a,j,f,d in zip(antResponse, flux, freq, cblDelay):
                factor = a * numpy.sqrt(j)
                angle = 2*numpy.pi*f*(times + (d - (geoDelay-geoDelayPC)))
                temp[srcCount,antCount,:] += factor*(numpy.cos(angle) + 1j*numpy.sin(angle))
            antCount = antCount + 1
        srcCount = srcCount + 1

    # Sum over sources and done
    tdSignals = temp.sum(axis=0)
    return tdSignals


def __pointSourceTBW(fh, stands, src, nFrames, **kwargs):
    """
    Private function to build TBW point sources.
    """
    
    bits = kwargs['bits']
    tStart = kwargs['tStart']
    phaseCenter = kwargs['phaseCenter']
    verbose = kwargs['verbose']
    noiseStrength = kwargs['noiseStrength']
    
    sampleRate = dp_common.fS
    freqs = (numpy.fft.fftfreq(1024, d=1.0/sampleRate))[1:512]
    aa = __getAntennaArray(lwa_common.lwa1, stands, tStart, freqs)
    
    if bits == 12:
        maxValue = 2047
        samplesPerFrame = 400
    else:
        maxValue =  7
        samplesPerFrame = 1200

    if verbose:
        print("Simulating %i captures of %i-bit TBW data for %i stands:" % \
            (int(numpy.ceil(nFrames / 30000.0)), bits, len(stands)))

    nCaptures = int(numpy.ceil(nFrames / 30000.0))
    for capture in range(nCaptures):
        j = 0
        k = 1
        for stand1, stand2 in zip(stands[0::2], stands[1::2]):
            FramesThisBatch = nFrames - capture*30000
            if FramesThisBatch > 30000:
                FramesThisBatch = 30000
            if verbose:
                print(" capture %i, stands %i and %i" % (capture+1, stand1, stand2))

            for i in range(FramesThisBatch):
                t = long(tStart*dp_common.fS) + i*samplesPerFrame
                t += long(60*dp_common.fS*capture)
                tFrame = t/dp_common.fS - tStart + numpy.arange(samplesPerFrame, dtype=numpy.float32) / dp_common.fS

                # Get the source parameters
                srcParams = __getSourceParameters(aa, tFrame[0], src)

                # Generate the time series response of each signal at each frequency
                tdSignalsX = __buildSignals(aa, stands, srcParams, tFrame*1e9, pol='x', phaseCenter=phaseCenter)
                tdSignalsY = __buildSignals(aa, stands, srcParams, tFrame*1e9, pol='y', phaseCenter=phaseCenter)

                cFrame = tbw.SimFrame(stand=stand1.stand.id, frameCount=i+1, dataBits=bits, obsTime=t)
                cFrame.xy = numpy.random.randn(2, samplesPerFrame)
                cFrame.xy *= maxValue*noiseStrength
                cFrame.xy[0,:] += maxValue*tdSignalsX.real[j,:]
                cFrame.xy[1,:] += maxValue*tdSignalsY.real[j,:]
                
                cFrame.writeRawFrame(fh)

                cFrame = tbw.SimFrame(stand=stand2.stand.id, frameCount=i+1, dataBits=bits, obsTime=t)
                cFrame.xy = numpy.random.randn(2, samplesPerFrame)
                cFrame.xy *= maxValue*noiseStrength
                cFrame.xy[0,:] += maxValue*tdSignalsX.real[k,:]
                cFrame.xy[1,:] += maxValue*tdSignalsY.real[k,:]
                
                cFrame.writeRawFrame(fh)


def __pointSourceTBN(fh, stands, src, nFrames, **kwargs):
    """
    Private function to build TBN point sources.
    """
    
    CentralFreq = kwargs['CentralFreq']
    filter = kwargs['filter']
    tStart = kwargs['tStart']
    phaseCenter = kwargs['phaseCenter']
    verbose = kwargs['verbose']
    noiseStrength = kwargs['noiseStrength']
    
    sampleRate = TBNFilters[filter]
    maxValue = 127
    samplesPerFrame = 512
    freqs = (numpy.fft.fftfreq(samplesPerFrame, d=1.0/sampleRate)) + CentralFreq
    freqs = numpy.fft.fftshift(freqs)
    aa = __getAntennaArray(lwa_common.lwa1, stands, tStart, freqs)
    
    if verbose:
        print("Simulating %i frames of TBN Data @ %.2f kHz for %i stands:" % \
            (nFrames, sampleRate/1e3, len(stands)))
    
    for i in range(nFrames):
        if i % 1000 == 0 and verbose:
            print(" frame %i" % (i+1))
        t = long(tStart*dp_common.fS) + long(i*dp_common.fS*samplesPerFrame/sampleRate)
        tFrame = t/dp_common.fS - tStart + numpy.arange(samplesPerFrame, dtype=numpy.float32) / sampleRate
        
        # Get the source parameters
        srcParams = __getSourceParameters(aa, tFrame[0], src)
        
        # Generate the time series response of each signal at each frequency
        tdSignalsX = __buildSignals(aa, stands, srcParams, tFrame*1e9, pol='x', phaseCenter=phaseCenter)
        tdSignalsY = __buildSignals(aa, stands, srcParams, tFrame*1e9, pol='y', phaseCenter=phaseCenter)
        
        j = 0
        for stand in stands:
            cFrame = tbn.SimFrame(stand=stand.stand.id, pol=0, centralFreq=CentralFreq, gain=19, frameCount=i+1, obsTime=t)
            cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
            cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
            cFrame.iq *= maxValue*noiseStrength
            cFrame.iq += maxValue*tdSignalsX[j,:].astype(numpy.singlecomplex)
            
            cFrame.writeRawFrame(fh)

            cFrame = tbn.SimFrame(stand=stand.stand.id, pol=1, centralFreq=CentralFreq, gain=19, frameCount=i+1, obsTime=t)
            cFrame.iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
            cFrame.iq += numpy.random.randn(samplesPerFrame) + 1j*numpy.random.randn(samplesPerFrame)
            cFrame.iq *= maxValue*noiseStrength
            cFrame.iq += maxValue*tdSignalsY[j,:].astype(numpy.singlecomplex)
            
            cFrame.writeRawFrame(fh)
            
            j += 1


def pointSource(fh, stands, src, nFrames, mode='TBN', CentralFreq=49.0e6, filter=7, bits=12, tStart=0, phaseCenter='z', noiseStrength=0.1, verbose=False):
    """
    Generate a collection of frames with a point source signal for TBW
    and TBN.  The point source is specified as a aipy.src object.
        
    All modes need to have stands (beams in the case of DRX) and number of
    frames to generate.  TBW also needs to `bits' keyword set to generate 
    either 12-bit or 4-bit data.  The TBN frames need the `filter' keyword 
    set to specify the filter width.
    
    .. versionchanged:: 0.4.4
        Added the `noiseStrength` keyword to control how much noise is added to 
        the data.
    """

    if tStart == 0:
        tStart = time.time()

    if mode == 'TBW':
        __pointSourceTBW(fh, stands, src, nFrames, bits=bits, tStart=tStart, phaseCenter=phaseCenter, noiseStrength=noiseStrength, verbose=verbose)
    elif mode == 'TBN':
        __pointSourceTBN(fh, stands, src, nFrames, CentralFreq=CentralFreq, filter=filter, tStart=tStart, phaseCenter=phaseCenter, noiseStrength=noiseStrength, verbose=verbose)
    else:
        raise RuntimeError("Unknown observations mode: %s" % mode)



# -*- coding: utf-8 -*-

"""
Python module to read in DRX data.  This module defines the following 
classes for storing the DRX data found in a file:

Frame
  object that contains all data associated with a particular DRX frame.  
  The primary constituents of each frame are:
    * FrameHeader - the DRX frame header object and
    * FrameData   - the DRX frame data object.
Combined, these two objects contain all of the information found in the 
original DRX frame.

ObservingBlock
object that stores a collection of Frames for all beams/tunings/
polarizations for a particular time.

The functions defined in this module fall into two class:
  1. convert a frame in a file to a Frame object and
  2. describe the format of the data in the file.

For reading in data, use the readFrame function.  It takes a python file-
handle as an input and returns a fully-filled Frame object.  The readBlock
function reads in a (user-defined) number of DRX frames and returns a 
ObservingBlock object.

For describing the format of data in the file, two function are provided:

getBeamCount
  read in the first few frames of an open file handle and return how many 
  beams are present in the file.

getFramesPerObs
  read in the first several frames to see how many frames (tunings/polarizations)
  are associated with each beam.

..versionchanged:: 1.2.0
    Dropped support for ObservingBlock since the lsl.reader.buffer modules does
    a better job.

.. versionchanged:: 0.4.0
    Switched over from pure Python readers to the new C-base Go Fast! readers.
"""

import copy
import numpy

from lsl.common import dp as dp_common
from lsl.reader._gofast import readDRX
from lsl.reader._gofast import syncError as gsyncError
from lsl.reader._gofast import eofError as geofError
from lsl.reader.errors import baseReaderError, syncError, eofError

__version__ = '0.8'
__revision__ = '$Rev: 2465 $'
__all__ = ['FrameHeader', 'FrameData', 'Frame', 'readFrame', 
           'getSampleRate', 'getBeamCount', 'getFramesPerObs', 'FrameSize', 'filterCodes', 
           '__version__', '__revision__', '__all__']

FrameSize = 4128

# List of filter codes and their corresponding sample rates in Hz
# NOTE: filter code 7 is only valid on DP DRX
filterCodes = {1: 250000, 2: 500000, 3: 1000000, 4: 2000000, 5: 4900000, 6: 9800000, 7: 19600000}


class FrameHeader(object):
    """
    Class that stores the information found in the header of a DRX 
    frame.  All six fields listed in the DP ICD version H are stored as 
    well as the original binary header data.
    """
    
    def __init__(self, frameCount=None, drxID=None, secondsCount=None, decimation=None, timeOffset=None):
        self.frameCount = frameCount
        self.drxID = drxID
        self.secondsCount = secondsCount
        self.decimation = decimation
        self.timeOffset = timeOffset
    
    def parseID(self):
        """
        Parse the DRX ID into a tuple containing the beam (1 through
        4), tunning (1 and 2), and polarization (0 and 1).
        """
        
        beam = self.drxID&7
        tune = (self.drxID>>3)&7
        pol  = (self.drxID>>7)&1

        return (beam, tune, pol)
    
    def getSampleRate(self):
        """
        Return the sample rate of the data in samples/second.
        """
        
        sampleRate = dp_common.fS / self.decimation
        return sampleRate
        
    def getFilterCode(self):
        """
        Function to convert the sample rate in Hz to a filter code.
        """
        
        sampleCodes = {}
        for key in filterCodes:
            value = filterCodes[key]
            sampleCodes[value] = key

        return sampleCodes[self.getSampleRate()]


class FrameData(object):
    """
    Class that stores the information found in the data section of a DRX
    frame.  All three fields listed in the DP ICD version H are stored.
    """

    def __init__(self, timeTag=None, tuningWord=None, flags=None, iq=None):
        self.centralFreq = None
        self.gain = None
        self.timeTag = timeTag
        self.tuningWord = tuningWord
        self.flags = flags
        self.iq = iq
        
    def getCentralFreq(self):
        """
        Function to set the central frequency of the DRX data in Hz.
        """

        return dp_common.fS * self.tuningWord / 2**32

    def setGain(self, gain):
        """
        Function to set the gain of the DRX data.
        """

        self.gain = gain


class Frame(object):
    """
    Class that stores the information contained within a single DRX 
    frame.  It's properties are FrameHeader and FrameData objects.
    """

    def __init__(self, header=None, data=None):
        if header is None:
            self.header = FrameHeader()
        else:
            self.header = header
            
        if data is None:
            self.data = FrameData()
        else:
            self.data = data
            
        self.valid = True

    def parseID(self):
        """
        Convenience wrapper for the Frame.FrameHeader.parseID 
        function.
        """
        
        return self.header.parseID()

    def getSampleRate(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getSampleRate 
        function.
        """
        
        return self.header.getSampleRate()
        
    def getFilterCode(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getFilterCode function.
        """

        return self.header.getFilterCode()

    def getTime(self):
        """
        Function to convert the time tag from samples since the UNIX epoch
        (UTC 1970-01-01 00:00:00) to seconds since the UNIX epoch.
        """

        seconds = (self.data.timeTag - self.header.timeOffset) / dp_common.fS
        
        return seconds
    
    def getCentralFreq(self):
        """
        Convenience wrapper for the Frame.FrameData.getCentralFreq function.
        """

        return self.data.getCentralFreq()

    def setGain(self, gain):
        """
        Convenience wrapper for the Frame.FrameData.setGain function.
        """

        self.data.setGain(gain)

    def __add__(self, y):
        """
        Add the data sections of two frames together or add a number 
        to every element in the data section.
        """
    
        newFrame = copy.deepcopy(self)
        newFrame += y
        return newFrame
            
    def __iadd__(self, y):
        """
        In-place add the data sections of two frames together or add 
        a number to every element in the data section.
        """
        
        try:
            self.data.iq += y.data.iq
        except AttributeError:
            self.data.iq += numpy.complex64(y)
        return self
        
    def __mul__(self, y):
        """
        Multiple the data sections of two frames together or multiply 
        a number to every element in the data section.
        """
        
        newFrame = copy.deepcopy(self)
        newFrame *= y
        return newFrame
            
    def __imul__(self, y):
        """
        In-place multiple the data sections of two frames together or 
        multiply a number to every element in the data section.
        """
        
        try:
            self.data.iq *= y.data.iq
        except AttributeError:
            self.data.iq *= numpy.complex64(y)
        return self
            
    def __eq__(self, y):
        """
        Check if the time tags of two frames are equal or if the time
        tag is equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX == tY:
            return True
        else:
            return False
            
    def __ne__(self, y):
        """
        Check if the time tags of two frames are not equal or if the time
        tag is not equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX != tY:
            return True
        else:
            return False
            
    def __gt__(self, y):
        """
        Check if the time tag of the first frame is greater than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX > tY:
            return True
        else:
            return False
            
    def __ge__(self, y):
        """
        Check if the time tag of the first frame is greater than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX >= tY:
            return True
        else:
            return False
            
    def __lt__(self, y):
        """
        Check if the time tag of the first frame is less than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX < tY:
            return True
        else:
            return False
            
    def __le__(self, y):
        """
        Check if the time tag of the first frame is less than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX <= tY:
            return True
        else:
            return False
            
    def __cmp__(self, y):
        """
        Compare two frames based on the time tags.  This is helpful for 
        sorting things.
        """
        
        tX = self.data.timeTag
        tY = y.data.timeTag
        if tY > tX:
            return -1
        elif tX > tY:
            return 1
        else:
            return 0


def readFrame(filehandle, Gain=None, Verbose=False):
    """
    Function to read in a single DRX frame (header+data) and store the 
    contents as a Frame object.  This function wraps readerHeader and 
    readData.
    """
    
    # New Go Fast! (TM) method
    try:
        newFrame = readDRX(filehandle, Frame())
    except gsyncError:
        mark = filehandle.tell() - FrameSize
        raise syncError(location=mark)
    except geofError:
        raise eofError
    
    if Gain is not None:
        newFrame.setGain(Gain)
        
    return newFrame


def getSampleRate(filehandle, nFrames=None, FilterCode=False):
    """
    Find out what the sampling rate/filter code is from a single observations.  
    By default, the rate in Hz is returned.  However, the corresponding filter 
    code can be returned instead by setting the FilterCode keyword to true.
    
    This function is included to make easier to write code for TBN analysis and 
    modify it for DRX data.
    """

    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()

    # Read in one frame
    newFrame = readFrame(filehandle)
    
    # Return to the place in the file where we started
    filehandle.seek(fhStart)

    if not FilterCode:
        return newFrame.getSampleRate()
    else:
        return newFrame.getFilterCode()


def getBeamCount(filehandle):
    """
    Find out how many beams are present by examining the first 32 DRX
    records.  Return the number of beams found.
    """

    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()

    # Build up the list-of-lists that store ID codes and loop through 32
    # frames.  In each case, parse pull the DRX ID, extract the beam number, 
    # and append the DRX ID to the relevant beam array if it is not already 
    # there.
    beams = []
    for i in range(32):
        cFrame = readFrame(filehandle)
            
        cID = cFrame.header.drxID
        beam = cID&7
        if beam not in beams:
            beams.append(beam)
            
    # Return to the place in the file where we started
    filehandle.seek(fhStart)

    # Return the number of beams found
    return len(beams)


def getFramesPerObs(filehandle):
    """
    Find out how many frames are present per beam by examining the first 
    32 DRX records.  Return the number of frames per observations as a four-
    element tuple, one for each beam.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Build up the list-of-lists that store ID codes and loop through 32
    # frames.  In each case, parse pull the DRX ID, extract the beam number, 
    # and append the DRX ID to the relevant beam array if it is not already 
    # there.
    idCodes = [[], [], [], []]
    for i in range(32):
        cFrame = readFrame(filehandle)
        
        cID = cFrame.header.drxID
        beam = cID&7
        if cID not in idCodes[beam-1]:
            idCodes[beam-1].append(cID)
            
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Get the length of each beam list and return them as a tuple
    return (len(idCodes[0]), len(idCodes[1]), len(idCodes[2]), len(idCodes[3]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
LWA_Common.astro module informal unit test / demo: give LWA-1 observation data
at current system clock time

Moved out of the lsl.astro module into this script and updated for LWA-1
"""

import argparse

from lsl import astro
from lsl.common import stations


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('-s', '--site', type=str, default='LWA-1',
                        help='site name')
    args = parser.parse_args()
    
    # setup transform objects
    site = args.site.lower().replace('-', '')
    if site == 'lwa1':
        station = stations.lwa1
    elif site == 'lwasv':
        station = stations.lwasv
    elif site == 'ovrolwa':
        station = stations.lwa1
        station.name = 'OVRO-LWA'
        station.lat, station.long, station.elev = ('37.2397808', '-118.2816819', 1183.4839)
    else:
        raise RuntimeError("Unknown site name: %s" % site)
        
    nam = station.name
    lng = astro.deg_to_dms(station.long*180/3.14)
    lat = astro.deg_to_dms(station.lat*180/3.14)
    lwa_lnlat = astro.lnlat_posn(lng, lat)
    
    print('---------------------------------------------------------------')
    print('%s location' % nam)
    print('---------------------------------------------------------------')
    print('Longitude:         %s (%0.3f)' % (lng, lwa_lnlat.lng))
    print('Latitude:          %s (%0.3f)' % (lat, lwa_lnlat.lat)) 
        
    # calculate offset from GM
    
    lwa_gm_hms = astro.deg_to_hms(-lwa_lnlat.lng)
    lwa_gm_off = lwa_gm_hms.to_sec()
    print('GM Offset:         %s (%0.3f)' % (lwa_gm_hms, lwa_gm_off))
    
    # get current UTC time from system clock
    
    utc = astro.get_julian_from_sys()
    utcd = astro.get_date(utc)
    lcld = utcd.to_zone()
    unixt = astro.get_timet_from_julian(utc)
    
    # caculate sidereal times
    
    gm_sid = astro.get_apparent_sidereal_time(utc)
    lwa_sid = astro.get_local_sidereal_time(lwa_lnlat.lng, utc)
        
    print('---------------------------------------------------------------')
    print('Current time')
    print('---------------------------------------------------------------')
    print('UTC time:             %s (%0.3f)' % (utcd, utc))
    print('%s local time:     %s' % (nam, str(lcld)))
    print('GM sidereal time:     %0.3f' % gm_sid)
    print('%s sidereal time:  %0.3f' % (nam, lwa_sid))
    print('UNIX time:            %d' % unixt)   
    
    # calculate nutation
    
    nut = astro.get_nutation(utc)
    (nut_lng, nut_obl, nut_ecl) = nut.format()
    
    print('---------------------------------------------------------------')
    print('Nutation')
    print('---------------------------------------------------------------')
    print('Longitude nutation: %s (%0.4f)' % (nut_lng, nut.longitude))
    print('Obliquity nutation: %s (%0.4f)' % (nut_obl, nut.obliquity))
    print('Ecliptic obliquity: %s (%0.4f)' % (nut_ecl, nut.ecliptic))
    
    # calculate Solar phenomena
    
    sun_rst = astro.get_solar_rst(utc, lwa_lnlat)
    (sun_utc_rise, sun_utc_set, sun_utc_trans) = sun_rst.format()
    sun_lcl_rise = sun_utc_rise.to_zone()
    sun_lcl_trans = sun_utc_trans.to_zone()
    sun_lcl_set = sun_utc_set.to_zone()
    
    sun_equ = astro.get_solar_equ_coords(utc) 
    (sun_ra, sun_dec) = sun_equ.format()
    sun_hrz = sun_equ.to_hrz(lwa_lnlat, utc)
    sun_ecl = sun_equ.to_ecl(utc)
    (sun_lng, sun_lat) = sun_ecl.format()
    
    print('---------------------------------------------------------------')
    print('Sun')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (sun_ra, sun_equ.ra))
    print('DEC:               %s (%0.3f)' % (sun_dec, sun_equ.dec)) 
    print('Ecl longitude:     %s (%0.3f)' % (sun_lng, sun_ecl.lng))
    print('Ecl latitude:      %s (%0.3f)' % (sun_lat, sun_ecl.lat))             
    print('Rise:              %s (%0.3f) [%s]' % (sun_utc_rise, sun_rst.rise, sun_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (sun_utc_trans, sun_rst.transit, sun_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (sun_utc_set, sun_rst.set, sun_lcl_set))
    print('Azimuth:           %0.3f %s' % (sun_hrz.az, astro.hrz_to_nswe(sun_hrz)))
    print('Altitude:          %0.3f' % sun_hrz.alt)
    print('Zenith:            %0.3f' % sun_hrz.zen()) 
        
    
    # calculate Lunar phenomena
    
    moon_rst = astro.get_lunar_rst(utc, lwa_lnlat)
    (moon_utc_rise, moon_utc_set, moon_utc_trans) = moon_rst.format()
    moon_lcl_rise = moon_utc_rise.to_zone()
    moon_lcl_trans = moon_utc_trans.to_zone()
    moon_lcl_set = moon_utc_set.to_zone()
    
    moon_equ = astro.get_lunar_equ_coords(utc) 
    (moon_ra, moon_dec) = moon_equ.format()
    moon_hrz = moon_equ.to_hrz(lwa_lnlat, utc)
    moon_ecl = moon_equ.to_ecl(utc)
    (moon_lng, moon_lat) = moon_ecl.format()
    
    moon_sun_ang = sun_equ.angular_separation(moon_equ)
    
    print('---------------------------------------------------------------')
    print('Moon')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (moon_ra, moon_equ.ra))
    print('DEC:               %s (%0.3f)' % (moon_dec, moon_equ.dec)) 
    print('Ecl longitude:     %s (%0.3f)' % (moon_lng, moon_ecl.lng))
    print('Ecl latitude:      %s (%0.3f)' % (moon_lat, moon_ecl.lat))             
    print('Rise:              %s (%0.3f) [%s]' % (moon_utc_rise, moon_rst.rise, moon_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (moon_utc_trans, moon_rst.transit, moon_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (moon_utc_set, moon_rst.set, moon_lcl_set))
    print('Azimuth:           %0.3f %s' % (moon_hrz.az, astro.hrz_to_nswe(moon_hrz)))
    print('Altitude:          %0.3f' % moon_hrz.alt)
    print('Zenith:            %0.3f' % moon_hrz.zen())
    print('Sun angle:         %0.3f' % moon_sun_ang) 
    
    
    # calculate Venus phenomena
    
    venus_rst = astro.get_venus_rst(utc, lwa_lnlat)
    (venus_utc_rise, venus_utc_set, venus_utc_trans) = venus_rst.format()
    venus_lcl_rise = venus_utc_rise.to_zone()
    venus_lcl_trans = venus_utc_trans.to_zone()
    venus_lcl_set = venus_utc_set.to_zone()
    
    venus_equ = astro.get_venus_equ_coords(utc) 
    (venus_ra, venus_dec) = venus_equ.format()
    venus_hrz = venus_equ.to_hrz(lwa_lnlat, utc)
    venus_ecl = venus_equ.to_ecl(utc)
    (venus_lng, venus_lat) = venus_ecl.format()
    
    venus_sun_ang = sun_equ.angular_separation(venus_equ)
    venus_elong = sun_ecl.lng - venus_ecl.lng
    if venus_elong < 0:
        venus_dir = 'E'
    else:
        venus_dir = 'W'
    
    print('---------------------------------------------------------------')
    print('Venus')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (venus_ra, venus_equ.ra))
    print('DEC:               %s (%0.3f)' % (venus_dec, venus_equ.dec)) 
    print('Ecl longitude:     %s (%0.3f)' % (venus_lng, venus_ecl.lng))
    print('Ecl latitude:      %s (%0.3f)' % (venus_lat, venus_ecl.lat))             
    print('Rise:              %s (%0.3f) [%s]' % (venus_utc_rise, venus_rst.rise, venus_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (venus_utc_trans, venus_rst.transit, venus_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (venus_utc_set, venus_rst.set, venus_lcl_set))
    print('Azimuth:           %0.3f %s' % (venus_hrz.az, astro.hrz_to_nswe(venus_hrz)))
    print('Altitude:          %0.3f' % venus_hrz.alt)
    print('Zenith:            %0.3f' % venus_hrz.zen())
    print('Sun angle:         %0.3f' % venus_sun_ang)
    print('Elongation:        %s %s (%0.3f)' % (astro.deg_to_dms(venus_elong), venus_dir, venus_elong))
    
    
    # calculate Mars phenomena
    
    mars_rst = astro.get_mars_rst(utc, lwa_lnlat)
    (mars_utc_rise, mars_utc_set, mars_utc_trans) = mars_rst.format()
    mars_lcl_rise = mars_utc_rise.to_zone()
    mars_lcl_trans = mars_utc_trans.to_zone()
    mars_lcl_set = mars_utc_set.to_zone()
    
    mars_equ = astro.get_mars_equ_coords(utc) 
    (mars_ra, mars_dec) = mars_equ.format()
    mars_hrz = mars_equ.to_hrz(lwa_lnlat, utc)
    mars_ecl = mars_equ.to_ecl(utc)
    (mars_lng, mars_lat) = mars_ecl.format()
    
    mars_sun_ang = sun_equ.angular_separation(mars_equ)
    
    print('---------------------------------------------------------------')
    print('Mars')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (mars_ra, mars_equ.ra))
    print('DEC:               %s (%0.3f)' % (mars_dec, mars_equ.dec)) 
    print('Ecl longitude:     %s (%0.3f)' % (mars_lng, mars_ecl.lng))
    print('Ecl latitude:      %s (%0.3f)' % (mars_lat, mars_ecl.lat))             
    print('Rise:              %s (%0.3f) [%s]' % (mars_utc_rise, mars_rst.rise, mars_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (mars_utc_trans, mars_rst.transit, mars_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (mars_utc_set, mars_rst.set, mars_lcl_set))
    print('Azimuth:           %0.3f %s' % (mars_hrz.az, astro.hrz_to_nswe(mars_hrz)))
    print('Altitude:          %0.3f' % mars_hrz.alt)
    print('Zenith:            %0.3f' % mars_hrz.zen())
    print('Sun angle:         %0.3f' % mars_sun_ang)
    
    
    # calculate Jupiter phenomena
    
    jupiter_rst = astro.get_jupiter_rst(utc, lwa_lnlat)
    (jupiter_utc_rise, jupiter_utc_set, jupiter_utc_trans) = jupiter_rst.format()
    jupiter_lcl_rise = jupiter_utc_rise.to_zone()
    jupiter_lcl_trans = jupiter_utc_trans.to_zone()
    jupiter_lcl_set = jupiter_utc_set.to_zone()
    
    jupiter_equ = astro.get_jupiter_equ_coords(utc) 
    (jupiter_ra, jupiter_dec) = jupiter_equ.format()
    jupiter_hrz = jupiter_equ.to_hrz(lwa_lnlat, utc)
    jupiter_ecl = jupiter_equ.to_ecl(utc)
    (jupiter_lng, jupiter_lat) = jupiter_ecl.format()
    
    jupiter_sun_ang = sun_equ.angular_separation(jupiter_equ)
    
    print('---------------------------------------------------------------')
    print('Jupiter')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (jupiter_ra, jupiter_equ.ra))
    print('DEC:               %s (%0.3f)' % (jupiter_dec, jupiter_equ.dec)) 
    print('Ecl longitude:     %s (%0.3f)' % (jupiter_lng, jupiter_ecl.lng))
    print('Ecl latitude:      %s (%0.3f)' % (jupiter_lat, jupiter_ecl.lat))             
    print('Rise:              %s (%0.3f) [%s]' % (jupiter_utc_rise, jupiter_rst.rise, jupiter_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (jupiter_utc_trans, jupiter_rst.transit, jupiter_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (jupiter_utc_set, jupiter_rst.set, jupiter_lcl_set))
    print('Azimuth:           %0.3f %s' % (jupiter_hrz.az, astro.hrz_to_nswe(jupiter_hrz)))
    print('Altitude:          %0.3f' % jupiter_hrz.alt)
    print('Zenith:            %0.3f' % jupiter_hrz.zen())
    print('Sun angle:         %0.3f' % jupiter_sun_ang)
    
    
    # calculate Saturn phenomena
    
    saturn_rst = astro.get_saturn_rst(utc, lwa_lnlat)
    (saturn_utc_rise, saturn_utc_set, saturn_utc_trans) = saturn_rst.format()
    saturn_lcl_rise = saturn_utc_rise.to_zone()
    saturn_lcl_trans = saturn_utc_trans.to_zone()
    saturn_lcl_set = saturn_utc_set.to_zone()
    
    saturn_equ = astro.get_saturn_equ_coords(utc) 
    (saturn_ra, saturn_dec) = saturn_equ.format()
    saturn_hrz = saturn_equ.to_hrz(lwa_lnlat, utc)
    saturn_ecl = saturn_equ.to_ecl(utc)
    (saturn_lng, saturn_lat) = saturn_ecl.format()
    
    saturn_sun_ang = sun_equ.angular_separation(saturn_equ)
    
    print('---------------------------------------------------------------')
    print('Saturn')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (saturn_ra, saturn_equ.ra))
    print('DEC:               %s (%0.3f)' % (saturn_dec, saturn_equ.dec)) 
    print('Ecl longitude:     %s (%0.3f)' % (saturn_lng, saturn_ecl.lng))
    print('Ecl latitude:      %s (%0.3f)' % (saturn_lat, saturn_ecl.lat))             
    print('Rise:              %s (%0.3f) [%s]' % (saturn_utc_rise, saturn_rst.rise, saturn_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (saturn_utc_trans, saturn_rst.transit, saturn_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (saturn_utc_set, saturn_rst.set, saturn_lcl_set))
    print('Azimuth:           %0.3f %s' % (saturn_hrz.az, astro.hrz_to_nswe(saturn_hrz)))
    print('Altitude:          %0.3f' % saturn_hrz.alt)
    print('Zenith:            %0.3f' % saturn_hrz.zen())
    print('Sun angle:         %0.3f' % saturn_sun_ang)
    
    # calculate SgrA phenomena
    
    sgra_j2000_equ = astro.equ_posn(astro.hms(17, 42, 48.1), astro.dms(True, 28, 55, 8))
    sgra_equ = astro.get_apparent_posn(sgra_j2000_equ, utc)
    sgra_rst = astro.get_object_rst(utc, lwa_lnlat, sgra_equ)
    (sgra_utc_rise, sgra_utc_set, sgra_utc_trans) = sgra_rst.format()
    sgra_lcl_rise = sgra_utc_rise.to_zone()
    sgra_lcl_trans = sgra_utc_trans.to_zone()
    sgra_lcl_set = sgra_utc_set.to_zone()
    
    (sgra_ra, sgra_dec) = sgra_equ.format()
    sgra_hrz = sgra_equ.to_hrz(lwa_lnlat, utc)
    sgra_gal = sgra_equ.to_gal(utc)
    (sgra_l, sgra_b) = sgra_gal.format()
    
    sgra_sun_ang = sun_equ.angular_separation(sgra_equ)
    
    print('---------------------------------------------------------------')
    print('SgrA')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (sgra_ra, sgra_equ.ra))
    print('DEC:               %s (%0.3f)' % (sgra_dec, sgra_equ.dec)) 
    print('Gal longitude:     %s (%0.3f)' % (sgra_l, sgra_gal.l))
    print('Gal latitude:      %s (%0.3f)' % (sgra_b, sgra_gal.b))           
    print('Rise:              %s (%0.3f) [%s]' % (sgra_utc_rise, sgra_rst.rise, sgra_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (sgra_utc_trans, sgra_rst.transit, sgra_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (sgra_utc_set, sgra_rst.set, sgra_lcl_set))
    print('Azimuth:           %0.3f %s' % (sgra_hrz.az, astro.hrz_to_nswe(sgra_hrz)))
    print('Altitude:          %0.3f' % sgra_hrz.alt)
    print('Zenith:            %0.3f' % sgra_hrz.zen())
    print('Sun angle:         %0.3f' % sgra_sun_ang)
    
    # calculate CasA phenomena
    
    casa_j2000_equ = astro.equ_posn(astro.hms(23, 23, 22.7), astro.dms(False, 58, 49, 16))
    casa_equ = astro.get_apparent_posn(casa_j2000_equ, utc)
    
    (casa_ra, casa_dec) = casa_equ.format()
    casa_hrz = casa_equ.to_hrz(lwa_lnlat, utc)
    casa_gal = casa_equ.to_gal(utc)
    (casa_l, casa_b) = casa_gal.format()
    
    casa_sun_ang = sun_equ.angular_separation(casa_equ)
    
    print('---------------------------------------------------------------')
    print('CasA')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (casa_ra, casa_equ.ra))
    print('DEC:               %s (%0.3f)' % (casa_dec, casa_equ.dec)) 
    print('Gal longitude:     %s (%0.3f)' % (casa_l, casa_gal.l))
    print('Gal latitude:      %s (%0.3f)' % (casa_b, casa_gal.b))
    print('Rise:              Circumpolar')
    print('Transit:           -----------')
    print('Set:               Circumpolar')
    print('Azimuth:           %0.3f %s' % (casa_hrz.az, astro.hrz_to_nswe(casa_hrz)))
    print('Altitude:          %0.3f' % casa_hrz.alt)
    print('Zenith:            %0.3f' % casa_hrz.zen())
    print('Sun angle:         %0.3f' % casa_sun_ang)
        

    # calculate CygA phenomena
    
    cyga_j2000_equ = astro.equ_posn(astro.hms(19, 59, 27.8), astro.dms(False, 40, 44, 2))
    cyga_equ = astro.get_apparent_posn(cyga_j2000_equ, utc)
    cyga_rst = astro.get_object_rst(utc, lwa_lnlat, cyga_equ)
    (cyga_utc_rise, cyga_utc_set, cyga_utc_trans) = cyga_rst.format()
    cyga_lcl_rise = cyga_utc_rise.to_zone()
    cyga_lcl_trans = cyga_utc_trans.to_zone()
    cyga_lcl_set = cyga_utc_set.to_zone()
    
    (cyga_ra, cyga_dec) = cyga_equ.format()
    cyga_hrz = cyga_equ.to_hrz(lwa_lnlat, utc)
    cyga_gal = cyga_equ.to_gal(utc)
    (cyga_l, cyga_b) = cyga_gal.format()
    
    cyga_sun_ang = sun_equ.angular_separation(cyga_equ)
    
    print('---------------------------------------------------------------')
    print('CygA')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (cyga_ra, cyga_equ.ra))
    print('DEC:               %s (%0.3f)' % (cyga_dec, cyga_equ.dec)) 
    print('Gal longitude:     %s (%0.3f)' % (cyga_l, cyga_gal.l))
    print('Gal latitude:      %s (%0.3f)' % (cyga_b, cyga_gal.b))           
    print('Rise:              %s (%0.3f) [%s]' % (cyga_utc_rise, cyga_rst.rise, cyga_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (cyga_utc_trans, cyga_rst.transit, cyga_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (cyga_utc_set, cyga_rst.set, cyga_lcl_set))
    print('Azimuth:           %0.3f %s' % (cyga_hrz.az, astro.hrz_to_nswe(cyga_hrz)))
    print('Altitude:          %0.3f' % cyga_hrz.alt)
    print('Zenith:            %0.3f' % cyga_hrz.zen())
    print('Sun angle:         %0.3f' % cyga_sun_ang)
    
    
    # calculate TauA phenomena
    
    taua_j2000_equ = astro.equ_posn(astro.hms(5, 34, 32.0), astro.dms(False, 22, 00, 52.1))
    taua_equ = astro.get_apparent_posn(taua_j2000_equ, utc)
    taua_rst = astro.get_object_rst(utc, lwa_lnlat, taua_equ)
    (taua_utc_rise, taua_utc_set, taua_utc_trans) = taua_rst.format()
    taua_lcl_rise = taua_utc_rise.to_zone()
    taua_lcl_trans = taua_utc_trans.to_zone()
    taua_lcl_set = taua_utc_set.to_zone()
    
    (taua_ra, taua_dec) = taua_equ.format()
    taua_hrz = taua_equ.to_hrz(lwa_lnlat, utc)
    taua_gal = taua_equ.to_gal(utc)
    (taua_l, taua_b) = taua_gal.format()
    
    taua_sun_ang = sun_equ.angular_separation(taua_equ)
    
    print('---------------------------------------------------------------')
    print('TauA')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (taua_ra, taua_equ.ra))
    print('DEC:               %s (%0.3f)' % (taua_dec, taua_equ.dec)) 
    print('Gal longitude:     %s (%0.3f)' % (taua_l, taua_gal.l))
    print('Gal latitude:      %s (%0.3f)' % (taua_b, taua_gal.b))           
    print('Rise:              %s (%0.3f) [%s]' % (taua_utc_rise, taua_rst.rise, taua_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (taua_utc_trans, taua_rst.transit, taua_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (taua_utc_set, taua_rst.set, taua_lcl_set))
    print('Azimuth:           %0.3f %s' % (taua_hrz.az, astro.hrz_to_nswe(taua_hrz)))
    print('Altitude:          %0.3f' % taua_hrz.alt)
    print('Zenith:            %0.3f' % taua_hrz.zen())
    print('Sun angle:         %0.3f' % taua_sun_ang)

    # calculate B1919+21 phenomena
    
    b1919_j2000_equ = astro.equ_posn(astro.hms(19, 21, 44.80), astro.dms(False, 21, 53, 1.8))
    b1919_equ = astro.get_apparent_posn(b1919_j2000_equ, utc)
    b1919_rst = astro.get_object_rst(utc, lwa_lnlat, b1919_equ)
    (b1919_utc_rise, b1919_utc_set, b1919_utc_trans) = b1919_rst.format()
    b1919_lcl_rise = b1919_utc_rise.to_zone()
    b1919_lcl_trans = b1919_utc_trans.to_zone()
    b1919_lcl_set = b1919_utc_set.to_zone()
    
    (b1919_ra, b1919_dec) = b1919_equ.format()
    b1919_hrz = b1919_equ.to_hrz(lwa_lnlat, utc)
    b1919_gal = b1919_equ.to_gal(utc)
    (b1919_l, b1919_b) = b1919_gal.format()
    
    b1919_sun_ang = sun_equ.angular_separation(b1919_equ)
    
    print('---------------------------------------------------------------')
    print('B1919+21')
    print('---------------------------------------------------------------')
    print('RA:                %s (%0.3f)' % (b1919_ra, b1919_equ.ra))
    print('DEC:               %s (%0.3f)' % (b1919_dec, b1919_equ.dec)) 
    print('Gal longitude:     %s (%0.3f)' % (b1919_l, b1919_gal.l))
    print('Gal latitude:      %s (%0.3f)' % (b1919_b, b1919_gal.b))           
    print('Rise:              %s (%0.3f) [%s]' % (b1919_utc_rise, b1919_rst.rise, b1919_lcl_rise))
    print('Transit:           %s (%0.3f) [%s]' % (b1919_utc_trans, b1919_rst.transit, b1919_lcl_trans))
    print('Set:               %s (%0.3f) [%s]' % (b1919_utc_set, b1919_rst.set, b1919_lcl_set))
    print('Azimuth:           %0.3f %s' % (b1919_hrz.az, astro.hrz_to_nswe(b1919_hrz)))
    print('Altitude:          %0.3f' % b1919_hrz.alt)
    print('Zenith:            %0.3f' % b1919_hrz.zen())
    print('Sun angle:         %0.3f' % b1919_sun_ang)
    



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""Unit test for lsl.misc.mathutil module."""

import unittest
import logging
import math
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


import numpy
from scipy.special import sph_harm

from lsl.misc import mathutil


__revision__  = "$Rev: 2470 $"
__version__   = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class mathutil_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.misc.mathutil
    module."""
    
    def setUp(self):
        """Setup for unit tests."""
        
        # disable logger since we intentionally generate errors 
        logging.basicConfig(stream = StringIO())
    
    def test_downsample(self):
        """Test mathutil.downsample() function."""
        
        x = numpy.array((0, 1, 2, 3))
        
        y = mathutil.downsample(x, 2, False)
        self.assertEqual(len(y), 2)
        self.assertAlmostEqual(y[0], 1.0)
        self.assertAlmostEqual(y[1], 5.0)
        
        y = mathutil.downsample(x, 2, True)
        self.assertEqual(len(y), 2)
        self.assertAlmostEqual(y[0], 0.5)
        self.assertAlmostEqual(y[1], 2.5)
        
        y = mathutil.downsample(x, 3)
        self.assertEqual(len(y), 1)
        self.assertAlmostEqual(y[0], 1.0)
        
    def test_cmagnitde(self):
        """Test mathutil.cmagnitude() function."""
        
        x = complex(1, 0)
        self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
        
        x = complex(0, 1)
        self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
        
        x = complex(-1, 0)
        self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
        
        x = complex(0, -1)
        self.assertAlmostEqual(mathutil.cmagnitude(x), 1.0)
        
        x = complex(3, 4)
        self.assertAlmostEqual(mathutil.cmagnitude(x), 5.0)
        
    def test_cphase(self):
        """Test mathutil.cphase() function."""
        
        x = complex(1, 0)
        self.assertAlmostEqual(mathutil.cphase(x), math.radians(0.0))
        
        x = complex(0, 1)
        self.assertAlmostEqual(mathutil.cphase(x), math.radians(90.0))
        
        x = complex(-1, 0)
        self.assertAlmostEqual(mathutil.cphase(x), math.radians(180.0))
        
        x = complex(0, -1)
        self.assertAlmostEqual(mathutil.cphase(x), math.radians(-90.0))
        
    def test_cpolar(self):
        """Test mathutil.cpolar() function."""
        
        x = numpy.array((0+1j, 1+0j, 1+1j, -1+0j))
        mag = (1.0, 1.0, math.sqrt(2.0), 1.0)
        phase = (math.radians(90.0), 0.0, math.radians(45.0), math.radians(180.0))
        im = iter(mag)
        ip = iter(phase)
        
        pol = mathutil.cpolar(x)
        for p in pol:
            self.assertAlmostEqual(p[0], next(im))
            self.assertAlmostEqual(p[1], next(ip))
            
    def test_creal(self):
        """Test mathtutil.creal() function."""
        
        x = (1, math.radians(0.0))
        self.assertAlmostEqual(mathutil.creal(x), 1.0)
        
        x = (1, math.radians(90.0))
        self.assertAlmostEqual(mathutil.creal(x), 0.0)
        
        x = (1, math.radians(180.0))
        self.assertAlmostEqual(mathutil.creal(x), -1.0)
        
        x = (1, math.radians(-90.0))
        self.assertAlmostEqual(mathutil.creal(x), 0.0)
            
    def test_cimag(self):
        """Test mathtutil.cimag() function."""
        
        x = (1, math.radians(0.0))
        self.assertAlmostEqual(mathutil.cimag(x), 0.0)
        
        x = (1, math.radians(90.0))
        self.assertAlmostEqual(mathutil.cimag(x), 1.0)
        
        x = (1, math.radians(180.0))
        self.assertAlmostEqual(mathutil.cimag(x), 0.0)
        
        x = (1, math.radians(-90.0))
        self.assertAlmostEqual(mathutil.cimag(x), -1.0)
        
    def test_crect(self):
        """Test mathutil.crect() function."""
        
        x = numpy.array(((1.0, math.radians(90.0)),
                    (1.0, 0.0),
                    (math.sqrt(2.0), math.radians(45.0)),
                    (1.0, math.radians(180.0))))
        
        c = (0+1j, 1+0j, 1+1j, -1+0j)
        ic = iter(c)
        
        rect = mathutil.crect(x)
        for r in rect:
            co = next(ic)
            self.assertAlmostEqual(r.real, co.real)
            self.assertAlmostEqual(r.imag, co.imag)
            
    def test_regrid(self):
        """Test mathutil.regrid() function."""
        
        yout = (0.0, 2.0, 4.0, 6.0, 8.0)
        
        x = numpy.arange(0, 10, dtype = numpy.float32)
        y = numpy.arange(0, 10, dtype = numpy.float32)
        xnew = numpy.arange(0, 10, 2, dtype = numpy.float32)
        
        ynew = mathutil.regrid(x, y, xnew, method = 'spline')
        iy = iter(yout)
        for yn in ynew:
            self.assertAlmostEqual(yn, next(iy))
            
        ynew = mathutil.regrid(x, y, xnew, method = 'linear')
        iy = iter(yout)
        for yn in ynew:
            self.assertAlmostEqual(yn, next(iy))
        
        xnew = numpy.arange(-2, 10, 2, dtype = numpy.float32)    
        self.assertRaises(ValueError, mathutil.regrid, x, y, xnew)
        
        xnew = numpy.arange(0, 12, 2, dtype = numpy.float32)    
        self.assertRaises(ValueError, mathutil.regrid, x, y, xnew)
        
    def test_smooth(self):
        """Test mathutil.smooth() function."""
        
        x = numpy.arange(0, 100, dtype = numpy.float32)
        mathutil.smooth(x)
        mathutil.smooth(x, window='flat')
        
    def test_to_dB(self):
        """Test mathutil.to_dB() function."""
        
        x = numpy.random.randn(100) + 1000.0
        mathutil.to_dB(x)

    def test_from_dB(self):
        """Test mathutil.from_dB function."""
        
        x = numpy.arange(1, 100, dtype = numpy.float32)
        mathutil.from_dB(x)
        
    def test_savitzky_golay(self):
        """Test Savitzky-Golay smoothing."""
        
        # Actual smoothing
        data = numpy.random.rand(4096)
        out = mathutil.savitzky_golay(data, 15, 3)
        
        # Derivative
        out = mathutil.savitzky_golay(data, 15, 3, deriv=1)
        
    def test_gaussian_gen(self):
        """Test 1-D and 2-D Gaussisan generating functions."""
        
        # 1-D
        height = 1
        center = 5.0
        width = 2.1
        gauFnc = mathutil.gaussian1d(height, center, width)
        value = gauFnc(numpy.arange(0, 100))
        
        # 2-D
        centerX = center
        centerY = -centerX
        widthX = width
        widthY = widthX/2
        gauFnc = mathutil.gaussian2d(height, centerX, centerY, widthX, widthY)
        value = gauFnc(numpy.arange(0, 100), numpy.arange(0,100))
        
    def test_gaussian_par(self):
        """Test 1-D and 2-D Gaussisan parameter estimation."""
        
        # 1-D
        height = 1.5
        center = 50.0
        width = 2.1
        gauFnc = mathutil.gaussian1d(height, center, width)
        
        x = numpy.arange(0, 100)
        value = gauFnc(x)
        
        params = mathutil.gaussparams(value)
        params = mathutil.gaussparams(value, x=x)
        self.assertAlmostEqual(height, params[0], 1)
        self.assertAlmostEqual(center, params[1], 1)
        self.assertAlmostEqual(width,  params[2], 1)
        
        # 2-D
        centerX = center
        centerY = center - 20.0
        widthX = width
        widthY = widthX/2.0
        gauFnc = mathutil.gaussian2d(height, centerX, centerY, widthX, widthY)
        
        x = numpy.zeros((100,100))
        y = numpy.zeros_like(x)
        for i in range(100):
            x[i,:] = i
            y[:,i] = i
        value = gauFnc(x, y)
        
        params = mathutil.gaussparams(value)
        params = mathutil.gaussparams(value, x=x, y=y)
        self.assertAlmostEqual(height,  params[0], 1)
        self.assertAlmostEqual(centerX, params[1], 1)
        self.assertAlmostEqual(centerY, params[2], 1)
        self.assertAlmostEqual(widthX,  params[3], 1)
        self.assertAlmostEqual(widthY,  params[4], 1)
    
    def test_sphval(self):
        """Test that the sphval() function runs."""
        
        terms = numpy.array([1, 0.5, 0.4, 0.01, -0.02, -0.005])
        az  = numpy.zeros((180,45))
        alt = numpy.zeros((180,45))
        for i in xrange(180):
            az[i,:] = 2*i
        for i in xrange(45):
            alt[:,i] = 2*i
        
        out = mathutil.sphval(terms, az, alt, degrees=True, realOnly=True)
        
        # Make sure we have the right output shape
        self.assertEqual(out.shape, az.shape)
    
    def test_sphfit(self):
        """Test the sphfit() function."""
        
        az  = numpy.zeros((180,45))
        alt = numpy.zeros((180,45))
        for i in xrange(180):
            az[i,:] = 2*i
        for i in xrange(45):
            alt[:,i] = 2*i
        
        # Setup a nice, easy problem
        out = 10*sph_harm(-1, 2, az*numpy.pi/180.0, alt*numpy.pi/180.0 + numpy.pi/2)

        # Evaluate the fit
        terms = mathutil.sphfit(az, alt, out, lmax=2, degrees=True)
        
        # Make sure the term with the most power correspond to the 2,-1 mode
        terms = (numpy.abs(terms) / numpy.abs(terms).max())**2
        self.assertEqual(numpy.where(terms == 1)[0][0], 5)

    
class mathutil_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lwa_user.mathutil
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(mathutil_tests))        
        
        
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    long = int

"""
Python module for creating creating, validating, and writing simulated 
TBN frames to a file.
"""

import numpy

from lsl.common.dp import fS
from lsl.reader import tbn
from lsl.sim.errors import invalidStand, invalidPol, invalidDataSize, invalidDataType

__version__ = '0.3'
__revision__ = '$Rev: 2544 $'
__all__ = ['SimFrame', 'frame2frame', '__version__', '__revision__', '__all__']


def frame2frame(tbnFrame):
    """
    Convert a :class:`lsl.reader.tbn.Frame` object to a raw DP TBN frame.
    """

    # The raw frame
    rawFrame = numpy.zeros(tbn.FrameSize, dtype=numpy.uint8)

    # Part 1: The header
    ## Sync. words (0xDEC0DE5C)
    rawFrame[0] = 0xDE  # 222
    rawFrame[1] = 0xC0  # 192
    rawFrame[2] = 0xDE  # 222
    rawFrame[3] = 0x5C  #  92
    ## Frame count
    rawFrame[5] = (tbnFrame.header.frameCount>>16) & 255
    rawFrame[6] = (tbnFrame.header.frameCount>>8) & 255
    rawFrame[7] = tbnFrame.header.frameCount & 255
    ## Tuning word
    rawFrame[8] = (tbnFrame.header.tuningWord>>24) & 255
    rawFrame[9] = (tbnFrame.header.tuningWord>>16) & 255
    rawFrame[10] = (tbnFrame.header.tuningWord>>8) & 255
    rawFrame[11] = tbnFrame.header.tuningWord & 255
    ## TBN ID
    rawFrame[12] = (tbnFrame.header.tbnID>>8) & 255
    rawFrame[13] = tbnFrame.header.tbnID & 255
    ## Gain
    rawFrame[14] = (tbnFrame.header.gain>>8) & 255
    rawFrame[15] = tbnFrame.header.gain & 255
    
    # Part 2: The data
    ## Time tag
    rawFrame[16] = (tbnFrame.data.timeTag>>56) & 255
    rawFrame[17] = (tbnFrame.data.timeTag>>48) & 255
    rawFrame[18] = (tbnFrame.data.timeTag>>40) & 255
    rawFrame[19] = (tbnFrame.data.timeTag>>32) & 255
    rawFrame[20] = (tbnFrame.data.timeTag>>24) & 255
    rawFrame[21] = (tbnFrame.data.timeTag>>16) & 255
    rawFrame[22] = (tbnFrame.data.timeTag>>8) & 255
    rawFrame[23] = tbnFrame.data.timeTag & 255
    ## Data
    iq = tbnFrame.data.iq
    ### Round and convert to unsigned integers
    iq = numpy.round(iq)
    if iq.dtype == numpy.complex128:
        iq = iq.astype(numpy.complex64)
    iq = iq.view(numpy.float32)
    iq = iq.astype(numpy.int8)
    
    rawFrame[24:] = iq
    
    return rawFrame


class SimFrame(tbn.Frame):
    """
    tbn.SimFrame extends the :class:`lsl.reader.tbn.Frame` object to yield a method 
    for easily creating DP ICD-compliant raw TBN frames.  Frames created with
    this method can be written to a file via the methods writeRawFrame() function.
    """

    def __init__(self, stand=None, pol=None, centralFreq=None, gain=None, frameCount=None, obsTime=None, iq=None):
        """
        Given a list of parameters, build a tbn.SimFrame object.  The parameters
        needed are:
        * stand id (>0 & <259)
        * polarization (0 for x, or 1 for y)
        * central frequency of tuning in (Hz)
        * TBN gain
        * which frame number to create
        * observation time in samples at fS since the epoch
        * 1-D numpy array representing the frame I/Q (complex) data
        
        Not all of these parameters are needed at initialization of the object and
        the values can be added later.

        .. versionchanged:: 0.3.4
            obsTime now in samples at fS, not seconds

        .. versionchanged:: 0.5.0
            Added support for ECR 11 TBN headers
        """
        
        self.stand = stand
        self.pol = pol
        self.freq = centralFreq
        self.gain = gain
        self.frameCount = frameCount
        self.obsTime = obsTime
        self.iq = iq
        super(SimFrame, self).__init__()
        
    def __update(self):
        """
        Private function to use the object's parameter values to build up 
        a tbn.Frame-like object.
        """
        
        self.header.frameCount = self.frameCount
        self.header.tuningWord = long( round(self.freq/fS*2**32) )
        self.header.tbnID = 2*(self.stand-1) + self.pol + 1
        self.header.gain = self.gain
        
        self.data.timeTag = self.obsTime
        self.data.iq = self.iq
    
    def loadFrame(self, tbnFrame):
        """
        Populate the a tbn.SimFrame object with a pre-made frame.
        """
        
        self.header = tbnFrame.header
        self.data = tbnFrame.data
        
        # Back-fill the class' fields to make sure the object is consistent
        ## Header
        self.stand = self.header.parseID()[0]
        self.pol = self.header.parseID()[1]
        self.freq = self.header.getCentralFreq()
        self.gain = self.header.getGain()
        self.frameCount = self.header.frameCount
        ## Data
        self.obsTime = self.data.timeTag
        self.iq = self.data.iq
    
    def isValid(self, raiseErrors=False):
        """
        Check if simulated TBN frame is valid or not.  Valid frames return 
        True and invalid frames False.  If the 'raiseErrors' keyword is set, 
        isValid raises an error when a problem is encountered.
        """

        # Make sure we have the latest values
        self.__update()

        stand, pol = self.parseID()
        # Is the stand number reasonable?
        if stand == 0 or stand > 258:
            if raiseErrors:
                raise invalidStand()
            return False

        # Is the polarization reasonable?
        if pol not in [0, 1]:
            if raiseErrors:
                raise invalidPol()
            return False

        # Is there data loaded into frame.data.iq?
        if self.data.iq is None:
            if raiseErrors:
                raise invalidDataSize()
            return False

        # Does the data length make sense?
        if self.data.iq.shape[0] != 512:
            if raiseErrors:
                raise invalidDataSize()
            return False

        # Does the data type make sense?
        if self.data.iq.dtype.kind != 'c':
            if raiseErrors:
                raise invalidDataType()
            return False

        # If we made it this far, it's valid
        return True

    def createRawFrame(self):
        """
        Re-express a simulated TBN frame as a numpy array of unsigned 8-bit 
        integers.  Returns a numpy array if the frame  is valid.  If the frame 
        is not ICD-compliant, a errors.baseSimError-type error is raised.
        """

        # Make sure we have the latest values
        self.__update()

        self.isValid(raiseErrors=True)
        return frame2frame(self)

    def writeRawFrame(self, fh):
        """
        Write a simulated TBN frame to a filehandle if the frame is valid.
        If the frame is not ICD-compliant, a errors.baseSimError-type error 
        is raised.
        """

        rawFrame = self.createRawFrame()
        rawFrame.tofile(fh)

    def __str__(self):
        if self.stand is None:
            return "Empty TBN SimFrame object"
        else:
            return "TBN SimFrame for stand %i, pol. %i @ time %i" % (self.stand, self.pol, self.obsTime)



# -*- coding: utf-8 -*-

"""
Time and position transform objects.
"""


import collections
import copy
import datetime
import math
import abc

from lsl import astro
from lsl.misc.total_sorting import cmp_to_total

__version__ = '0.2'
__revision__ = '$Rev: 2383 $'
__author__ = "Unknown"
__maintainer__ = "Jayce Dowell"


@cmp_to_total
class Time(object):
    """
    Holds an absolute time value and can manipulate the value in 
    various representations.
    
    After creation, a particular time format and time system may
    be accessed using the appropriate instance member.  If marked
    with '(S)', the Time value may also be updated by setting the
    member to a new value.
    
    utc_jd (S)        - UTC standard Julian day
    utc_mjd (S)       - UTC modified Julian day
    utc_timet (S)     - UTC UNIX timet seconds
    utc_py_date (S)   - UTC python datetime.datetime object
    utc_ln_date (S)   - UTC libnova astro.date object
    utc_str           - UTC ISO8601 calendar string format
    
    tai_jd (S)        - TAI standard Julian day
    tai_mjd (S)       - TAI modified Julian day
    tai_timet (S)     - TAI UNIX timet seconds
    """

    # time format types

    FORMAT_LN_DATE    = 'LN_DATE'
    FORMAT_PY_DATE    = 'PY_DATE'
    FORMAT_STR        = 'STR'
    FORMAT_JD         = 'JD'
    FORMAT_MJD        = 'MJD'
    FORMAT_TIMET      = 'TIMET'
    
    known_formats = (FORMAT_LN_DATE, FORMAT_PY_DATE, FORMAT_STR, FORMAT_JD, 
    FORMAT_MJD, FORMAT_TIMET)
    
    # time system types
    TIMESYS_UTC       = 'UTC'
    TIMESYS_TAI       = 'TAI'
    
    known_timesys = (TIMESYS_UTC, TIMESYS_TAI)
    
    @classmethod
    def from_system(klass):
        """
        Factory method to create a Time instance from current system clock value.
        """
        
        return klass(astro.get_julian_from_sys(), klass.FORMAT_JD, klass.TIMESYS_UTC)
    
    def __init__(self, value, format = FORMAT_MJD, timesys = TIMESYS_UTC):
        """
        Create a Time instance, using 'value' as the initial time.
        
        'format' describes the type of 'value'
            Time.FORMAT_LN_DATE - libnova astro.date class calendar format
            Time.FORMAT_PY_DATE - python datetime.datetime class calendar format
            Time.FORMAT_STR     - ISO 8601 (YYYY-MM-DD hh:mm:ss.s) calendar format 
                            string 
            Time.FORMAT_JD      - standard julian day as float
            Time.FORMAT_MJD     - modified julian day as float
            Time.FORMAT_TIMET   - UNIX timet seconds as integer
        
        'timesys' defines the time system
            Time.TIMESYS_UTC  - UTC time system
            Time.TIMESYS_TAI  - TAI time system
        """
        
        # check parameters
        if format not in self.known_formats:
            raise ValueError("unknown format %s" % format)
            
        if timesys not in self.known_timesys:
            raise ValueError("unknown timesys %s" % timesys)
        
        # parse init value base on format type
        # time value is held internally as UTC JD float
        if format == self.FORMAT_LN_DATE:
            self.utc_ln_date = value
            
        elif format == self.FORMAT_PY_DATE:
            self.utc_py_date = value
            
        elif format == self.FORMAT_STR:
            if not isinstance(value, str):
                raise TypeError("value must be type str")
            d = astro.date()
            d.load(*value.split())
            self._time = d.to_jd()
            
        elif format == self.FORMAT_JD:
            self.utc_jd = value
            
        elif format == self.FORMAT_MJD:
            self.utc_mjd = value
            
        elif format == self.FORMAT_TIMET:
            self.utc_timet = value
            
        # convert internal time to UTC if an alternate timesys is specified
        if timesys == self.TIMESYS_TAI:
            self._time = astro.tai_to_utc(self._time)
        
    def __cmp__(self, other):
        """
        Determine ordering for two Time instances.
        """
        
        return cmp(self._time, other._time)
    
    def __hash__(self):
        """
        Return a hash key for the Time instance.
        """
        
        return hash(self._time)
    
    def __repr__(self):
        """
        Return low-level string representation of Time instance.
        """
        
        return "%s.%s(%s)" % (type(self).__module__, type(self).__name__, repr(self.utc_mjd))
        
    def __str__(self):
        """
        Return high-level string representation of Time instance.
        """
        
        return self.utc_str
        
    @property
    def utc_jd(self):
        """
        Time value formatted as UTC standard julian day (float).
        """
        
        return self._time
        
    @utc_jd.setter
    def utc_jd(self, value):

        if not isinstance(value, (int, float)):
            raise TypeError("value must be type int or float")
            
        self._time = float(value)
        
    @property
    def utc_mjd(self):
        """
        Time value formatted as UTC modified julian day (float).
        """
        
        return astro.jd_to_mjd(self._time) 
        
    @utc_mjd.setter
    def utc_mjd(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("value must be type int or float")
            
        self._time = astro.mjd_to_jd(float(value))
        
    @property
    def utc_ln_date(self):
        """
        Time value formatted as UTC calendar astro.date object.
        """
        
        return astro.get_date(self._time)
        
    @utc_ln_date.setter
    def utc_ln_date(self, value):
        if not isinstance(value, astro.date):
            raise TypeError("value must be type astro.date")
            
        self._time = astro.get_julian_day(value)
        
    @property
    def utc_py_date(self):
        """
        Time value formattes as UTC calendar datetime.datetime object.
        """
        
        return self.date_ln_to_py(self.utc_ln_date)
        
    @utc_py_date.setter
    def utc_py_date(self, value):
        if not isinstance(value, datetime.datetime):
            raise ValueError("value must be type datetime.datetime")
            
        self.utc_ln_date = self.date_py_to_ln(value)
        
    @property
    def utc_str(self):
        """
        Time value formatted as UTC ISO 8601 calendar string.
        """
        
        return str(self.utc_ln_date)
        
    @property
    def tai_jd(self):
        """
        Time value formatted as TAI standard julian day (float).
        """
        
        return astro.utc_to_tai(self._time)
        
    @tai_jd.setter
    def tai_jd(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("value must be type int or float")
            
        self._time = astro.tai_to_utc(float(value))
        
    @property
    def tai_mjd(self):
        """
        Time value formatted as TAI modified julian day (float).
        """
        
        return astro.utcjd_to_taimjd(self._time)
        
    @tai_mjd.setter
    def tai_mjd(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("value must be type int or float")
            
        self._time = astro.taimjd_to_utcjd(float(value))
        
    @property
    def utc_timet(self):
        """
        Time value formatted as UTC UNIX timet seconds.
        """
        
        return astro.get_timet_from_julian(self._time)
        
    @utc_timet.setter
    def utc_timet(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("value must be type int or float")
            
        self._time = astro.get_julian_from_timet(int(value))
        
    @property
    def tai_timet(self):
        """
        Time value formatted as TAI UNIX timet seconds.
        """
        
        return astro.get_timet_from_julian(astro.utc_to_tai(self._time))
        
    @tai_timet.setter
    def tai_timet(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("value must be type int or float")
            
        self._time = astro.tai_to_utc(astro.get_julian_from_timet(int(value)))
        
    @staticmethod
    def date_py_to_ln(pyDate):
        """
        Convert python datatime.datetime object into a libnova astro.date
        object.
        """
        
        lnDate = astro.date(pyDate.year, pyDate.month, pyDate.day, pyDate.hour, pyDate.minute)
        lnDate.seconds = float(pyDate.second) + (pyDate.microsecond * 1e-6)
        return lnDate
        
    @staticmethod
    def date_ln_to_py(lnDate):
        """
        Convert libnova astro.date object into a python datetime.datetime
        object.
        """
        
        (usec, sec) = math.modf(lnDate.seconds)
        pyDate = datetime.datetime(lnDate.years, lnDate.months, lnDate.days,
            lnDate.hours, lnDate.minutes, int(sec), int(usec * 1e6))
        return pyDate


class SkyPosition(object):
    """
    Base abstract class for representing positions on the sky.
    """
    
    __metaclass__ = abc.ABCMeta
    
    @abc.abstractmethod
    def apparent_equ(self, time_):
        """
        Return position formatted as apparent equatorial coordinates.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return value is object of type astro.equ_posn.
        """
        
        return None
        
    def apparent_ecl(self, time_):
        """
        Return position formatted as apparent ecliptic coordinates.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return alue is object of type astro.ecl_posn.
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be of type transform.Time")
            
        equ = astro.get_apparent_posn(self._posn, time_.utc_jd)
        return astro.get_ecl_from_equ(equ, time_.utc_jd)


class CelestialPosition(SkyPosition):
    """
    Holds a celestial object position value and can manipulate
    the value in various representations.
    
    After creation, the celestial coordinates may be accessed in different
    formats and epoch by using the appropriate instance member.  If marked
    with '(S)', the Time value may also be updated by setting the member to 
    a new value. 
    
    j2000_equ (S)     - J2000 equatorial coordinates
    j2000_gal (S)     - J2000 galactic coordinates
    j2000_ecl (S)     - J2000 ecliptic coordinates
    
    b1950_equ (S)     - B1950 equatorial coordinates
    
    The instance methods apparent_equ() and apparent_ecl() may be used to
    get the apparent equatorial or ecliptic coordinates for a particular
    time.
    """
    
    # format types
    FORMAT_EQU      = 'EQU'
    FORMAT_GAL      = 'GAL'
    FORMAT_ECL      = 'ECL'
    
    known_formats = (FORMAT_EQU, FORMAT_ECL, FORMAT_GAL)
    
    # epoch types
    EPOCH_J2000     = 'J2000'
    EPOCH_B1950     = 'B1950'
    
    known_epochs = (EPOCH_J2000, EPOCH_B1950)
    
    def __init__(self, value, format = FORMAT_EQU, epoch = EPOCH_J2000, name = ''):
        """
        Create a CelestialPosition object, using 'value' as the initial coordinates. 
        
        'format' describes the type of 'value'
        * CelestialPosition.FORMAT_EQU  - equatorial coordinates RA and DEC
        * CelestialPosition.FORMAT_GAL  - galactic coordinates l and b
        * CelestialPosition.FORMAT_ECL  - ecliptic longitude and latitude
        
        'ephoch' defines the coordinate system equinox epoch
        * CelestialPosition.EPOCH_J2000   - J2000 epoch
        * CelestialPosition.EPOCH_B1950   - B1950 epoch 
        
        For format type EQU, the 'value' parameter may be either an instance of 
        the astro.equ_posn class or a sequence of length 2, in which case 
        RA=value[0] and DEC=value[1].
        
        For format type GAL, the 'value' parameter may be either an instance of
        the astro.gal_posn class or a sequence of length 2, in which case
        l=value[0] and b=value[1].
        
        For format type ECL, the 'value' parameter may be either an instance of
        the astro.ecl_posn class or a sequence of length 2, in which case
        long=value[0], lat=value[1]. 
        """
        
        # check parameters
        if format not in self.known_formats:
            raise ValueError("unknown format %s" % format)
            
        if epoch not in self.known_epochs:
            raise ValueError("unknown epoch %s" % epoch)
            
        self.name = name
        
        # parse init value based on format type
        # position is held internally as J2000 astro.equ_posn
        
        if format == self.FORMAT_EQU:
            if epoch == self.EPOCH_J2000:
                self.j2000_equ = value
            elif epoch == self.EPOCH_B1950:
                self.b1950_equ = value
                
        elif format == self.FORMAT_GAL:
            if epoch == self.EPOCH_J2000:
                self.j2000_gal = value
            else:
                raise ValueError("epoch %s not supported for GAL format" % epoch)
        elif format == self.FORMAT_ECL:
            if epoch == self.EPOCH_J2000:
                self.j2000_ecl = value
            else:
                raise ValueError("epoch %s not supported for ECL format" % epoch)
                
    def __repr__(self):
        return "%s.%s(%s, name=%s)" % (type(self).__module__, type(self).__name__, repr(self._posn), repr(self.name))
        
    @property
    def j2000_equ(self):
        """
        Position formatted as J2000 equatorial coordinates.
        Value is object of type astro.equ_posn.
        """
        
        return self._posn
        
    @j2000_equ.setter
    def j2000_equ(self, value):
        if not isinstance(value, (astro.equ_posn, collections.Sequence)):
            raise TypeError("value must be type astro.equ_posn or sequence of length 2")
        if isinstance(value, collections.Sequence):
            if len(value) != 2:
                raise TypeError("value sequence must be length 2")
            value = astro.equ_posn(*value)
            
        self._posn = copy.copy(value)
        
    def apparent_equ(self, time_):
        """
        Return position formatted as apparent equatorial coordinates.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return value is object of type astro.equ_posn.
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be of type transform.Time")
            
        return astro.get_apparent_posn(self._posn, time_.utc_jd)
        
    @property
    def b1950_equ(self):
        """
        Position formatted as B1950 equatorial coordinates.
        Value is object of type astro.equ_posn.
        """
        
        return astro.get_equ_prec2(self._posn, astro.J2000_UTC_JD, astro.B1950_UTC_JD)
    
    
    @b1950_equ.setter
    def b1950_equ(self, value):
        if not isinstance(value, (astro.equ_posn, collections.Sequence)):
            raise TypeError("value must be type astro.equ_posn or sequence of length 2")
        if isinstance(value, collections.Sequence):
            if len(value) != 2:
                raise TypeError("value sequence must be length 2")
            value = astro.equ_posn(*value)
        
        self._posn = astro.get_equ_prec2(value, astro.B1950_UTC_JD, astro.J2000_UTC_JD)
        
    @property
    def j2000_ecl(self):
        """
        Position formatted as J2000 ecliptic coordinates.
        Value is object of type astro.ecl_posn.
        """
        
        return astro.get_ecl_from_equ(self._posn, astro.J2000_UTC_JD)
        
    @j2000_ecl.setter
    def j2000_ecl(self, value):
        if not isinstance(value, (astro.equ_posn, collections.Sequence)):
            raise TypeError("value must be type astro.ecl_posn or sequence of length 2")
        if isinstance(value, collections.Sequence):
            if len(value) != 2:
                raise TypeError("value sequence must be length 2")
            value = astro.ecl_posn(*value)
            
        self._posn = astro.get_equ_from_ecl(value, astro.J2000_UTC_JD)
        
    def apparent_ecl(self, time_):
        """
        Return position formatted as apparent ecliptic coordinates.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return alue is object of type astro.ecl_posn.
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be of type transform.Time")
            
        equ = astro.get_apparent_posn(self._posn, time_.utc_jd)
        return astro.get_ecl_from_equ(equ, time_.utc_jd)
        
    @property
    def j2000_gal(self):
        """
        Position formatted as J2000 galactic coordinates.
        Value is object of type astro.gal_posn.
        """
        
        return astro.get_gal_from_equ2000(self._posn)
        
    @j2000_gal.setter
    def j2000_gal(self, value):
        if not isinstance(value, (astro.gal_posn, collections.Sequence)):
            raise TypeError("value must be type astro.gal_posn or sequence of length 2")
        if isinstance(value, collections.Sequence):
            if len(value) != 2:
                raise TypeError("value sequence must be length 2")
            value = astro.gal_posn(*value)
            
        self._posn = astro.get_equ2000_from_gal(value)


class PlanetaryPosition(SkyPosition):
    """
    Holds a solar, lunar, or planetary position value and can manipulate
    the value in various representations.
    
    The instance methods apparent_equ() and apparent_ecl() may be used to
    get the apparent equatorial or ecliptic coordinates for a particular
    time.
    """
    
    # planetary names
    NAME_SUN      = 'Sun'
    NAME_MOON     = 'Moon'
    NAME_VENUS    = 'Venus'
    NAME_MARS     = 'Mars'
    NAME_JUPITER  = 'Jupiter'
    NAME_SATURN   = 'Saturn'
    
    known_names = (NAME_SUN, NAME_MOON, NAME_VENUS, NAME_MARS, NAME_JUPITER, NAME_SATURN)
    
    def __init__(self, name):
        """
        Create a PlanetaryPosition instance.
        """
        
        # check parameters
        if name not in self.known_names:
            raise ValueError("unknown name %s" % name)
            
        self.name = name
        
        # cache the appropriate ephemeris function
        if name == self.NAME_SUN:
            self._posn_func = astro.get_solar_equ_coords
        elif name == self.NAME_MOON:
            self._posn_func = astro.get_lunar_equ_coords
        elif name == self.NAME_VENUS:
            self._posn_func = astro.get_venus_equ_coords
        elif name == self.NAME_MARS:
            self._posn_func = astro.get_mars_equ_coords
        elif name == self.NAME_JUPITER:
            self._posn_func = astro.get_jupiter_equ_coords
        elif name == self.NAME_SATURN:
            self._posn_func = astro.get_saturn_equ_coords
            
    def __repr__(self):
        return "%s.%s(%s)" % (type(self).__module__, type(self).__name__, repr(self.name))
        
    def apparent_equ(self, time_):
        """
        Position formatted as apparent equatorial coordinates.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return value is object of type astro.equ_posn.
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be of type transform.Time")
        
        return self._posn_func(time_.utc_jd)
        
    def apparent_ecl(self, time_):
        """
        Position formatted as apparent ecliptic coordinates.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return value is object of type astro.ecl_posn.
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be of type transform.Time")
        
        equ = self._posn_func(time_.utc_jd)
        return astro.get_ecl_from_equ(equ, time_.utc_jd)


class GeographicalPosition(object):
    """
    Holds a geographical position valuee and can manipulate the value in
    various representations.
    
    After creation, the celestial coordinates may be accessed in different
    formats and epoch by using the appropriate instance member.  If marked
    with '(S)', the Time value may also be updated by setting the member to 
    a new value.
    
    geo (S)     - longitude and latitude degrees, elevation meters
    ecef (S)    - Earth centered rectilinear coordinates
    """
    
    # format types
    FORMAT_GEO  = 'GEO'
    FORMAT_ECEF = 'ECEF'
    
    known_formats = (FORMAT_GEO, FORMAT_ECEF)
    
    def __init__(self, value, format = FORMAT_GEO, name = ''):
        """
        Create a new GeographicalPosition instance, using 'value' as the
        initial coordinates.
        
        'format' describes the type of 'value':
        * GeographicalPoistion.FORMAT_GEO   - longitude,latitude, and elevation
        * GeographicalPoistion.FORMAT_ECEF  - ECEF rectangular geodetic
        """
        
        # check parameters
        if format not in self.known_formats:
            raise ValueError("unknown format %s" % format) 
            
        self.name = name
        
        # parse init value based on format type
        # value is held internally as type astro.geo_posn
        if format == self.FORMAT_GEO:
            self.geo = value
            
        elif format == self.FORMAT_ECEF:
            self.ecef = value
            
    def __repr__(self):
        return "%s.%s(%s, name=%s)" % (type(self).__module__, type(self).__name__, repr(self._posn), repr(self.name))
        
    @property
    def geo(self):
        """
        Position formatted as geodedic longitude, latitude, elevation.
        Value is object of type astro.geo_posn.
        """
        
        return self._posn
        
    @geo.setter
    def geo(self, value):
        if not isinstance(value, (astro.geo_posn, collections.Sequence)):
            raise TypeError("value must be type astro.geo_posn or sequence of length 2/3")
        if isinstance(value, collections.Sequence):
            if (len(value) != 2) and (len(value) != 3):
                raise TypeError("value sequence must be length 2/3")
            value = astro.geo_posn(*value)
            
        self._posn = copy.copy(value)
        
    @property
    def ecef(self):
        """
        Position formatted as ECEF rectagular coordinates.
        Value is object of type astro.rect_posn.
        """
        
        return astro.get_rect_from_geo(self._posn)
        
    @ecef.setter
    def ecef(self, value):
        if not isinstance(value, (astro.rect_posn, collections.Sequence)):
            raise TypeError("value must be type astro.rect_posn or sequence of length 3")
        if isinstance(value, collections.Sequence):
            if len(value) != 3:
                raise TypeError("value sequence must be length 3")
            value = astro.rect_posn(*value)
            
        self._posn = astro.get_geo_from_rect(value)
        
    def sidereal(self, time_):
        """
        Return the apparent sidereal time for this location.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Returns sidereal time as a float in range [0.0, 24.0).
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be type transform.Time")
            
        return astro.get_local_sidereal_time(self._posn.lng, time_.utc_jd)


class PointingDirection(object):
    """
    Represent a target source object pointing direction for a particular ground
    site location.
    
    Each PointingDirection instance pairs a celestial position with a ground
    observation site position.  Each instance contains a 'source' member reference
    of type CelestialPosition or PlanetaryPosition and a 'site' member reference
    of type GeographicalPosition.
    
    The instance methods hrz() and dir_cos() may be called to get the
    pointing direction as horizontal coordinates or direction cosine components
    for a particular time of interest.  The rst() method may be called to get
    the rise, set, and transit ephemeris times.
    """
    
    def __init__(self, source, site):
        """
        Create a new pointing direction instance.
        
        source  - instance of type CelestialPosition or PlanetaryPosition 
                represeting target source location
        site    - instance of GeographicalPosition representing observer 
                location
        """
        
        # save parameters
        self.source = source
        self.site = site
        
    def __repr__(self):
        return "%s.%s(source=%s, site=%s)" % (type(self).__module__, type(self).__name__, repr(self.source), repr(self.site))
        
    def __setattr__(self, name, value):
        # make surce 'source' and 'site' member are correct type
        
        if (name == 'source') and (not isinstance(value, SkyPosition)):
            raise TypeError("\'source\' must be type SkyPosition")
            
        elif (name == 'site') and (not isinstance(value, GeographicalPosition)):
            raise TypeError("\'site\' must be type GeographicalPosition")
            
        object.__setattr__(self, name, value)
        
    def hrz(self, time_):
        """
        Return the pointing direction in horizontal coordinates as type 
        astro.hrz_posn.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be type transform.Time")
        
        return astro.get_hrz_from_equ(self.source.apparent_equ(time_), 
            self.site.geo, time_.utc_jd)
            
    def dir_cos(self, time_):
        """
        Return the pointing direction as three direction cosine components.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return value is a tuple (l, m, n) of direction cosine components.
        """
        return astro.dir_cos(self.hrz(time_))
        
    def rst(self, time_):
        """
        Return the rise, set, and transit ephemeris times.
        The 'time_' parameter should be set to a Time instance providing
        the time of interest.
        Return value is an object of type astro.rst_time, or None if
        the source is circumpolar.
        """
        
        if not isinstance(time_, Time):
            raise TypeError("time_ must be type transform.Time")
            
        return astro.get_object_rst(time_.utc_jd, self.site.geo, self.source.apparent_equ(time_))


# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function

"""
Module for removing multi-path scattering effects in pulsar profiles.  This 
is based on the CLEAN-like deconvolution method presented in Bhat, N., 
Cordes, J., & Chatterjee, S.  2003, ApJ, 584, 782.

http://iopscience.iop.org/0004-637X/584/2/782/fulltext/56392.text.html

.. note::
    All functions assume that the time and scattering time values are in
    seconds.
"""

import numpy
from lsl.statistics import robust


__version__ = "0.1"
__revision__ = "$Rev$"
__all__ = ['thin', 'thick', 'uniform', 'unscatter', '__version__', '__revision__', '__all__']


def thin(t, tauScatter):
    """
    Pulsar broadening function for multi-path scattering through a
    thin screen.
    """

    g  = 1.0/tauScatter * numpy.exp(-t/tauScatter)
    g  = numpy.where(t >= 0, g, 0)
    g /= g.sum()

    return g


def thick(t, tauScatter):
    """
    Pulse broadening function for multi-path scattering through a
    thick screen.

    From:
    Williamson, I. P. 1972, MNRAS, 157, 55; first equation on page 62 
    """

    tPrime = t + 1e-15
    
    g  = numpy.sqrt(numpy.pi*tauScatter/4/tPrime**3)
    g *= numpy.exp(-numpy.pi**2*tauScatter/16/tPrime)
    g  = numpy.where(t > 0, g, 0)
    g /= g.sum()

    return g


def uniform(t, tauScatter):
    """
    Pulsr broadening function for multi-path scattering through a 
    uniform screen.

    From:
    Williamson, I. P. 1972, MNRAS, 157, 55; last equation on page 65
    """

    tPrime = t + 1e-15
    
    g  = numpy.sqrt(numpy.pi**5*tauScatter**3/8/tPrime**5)
    g *= numpy.exp(-numpy.pi**2*tauScatter/4/tPrime)
    g  = numpy.where(t > 0, g, 0)
    g /= g.sum()

    return g


def _positivity(t, raw, resids, cc):
    """
    Compute the positivity of the residual profile following Equation 10 of
    Bhat, N., Cordes, J., & Chatterjee, S.  2003, ApJ, 
    584, 782.
    """
    
    # Weight factor
    m = 1.0

    # Tuning factor to find when the CLEANed profile goes negative
    x = 3./2.

    # Get the off-pulsar standard deviation
    sigma = robust.std(raw)
    
    temp = -resids - x*sigma
    f  = (resids**2 * numpy.where(temp>=0, 1, 0)).mean()
    f *= m/sigma**2

    return f

def _skewness(t, raw, resids, cc):
    """
    Compute the skewness for a collection of clean components following 
    Equation 12 of Bhat, N., Cordes, J., & Chatterjee, S.  2003, ApJ, 
    584, 782.
    """

    tM = (t*cc).sum() / cc.sum()
    t2 = ((t - tM)**2 * cc).sum()
    t3 = ((t - tM)**3 * cc).sum()
    
    return t3 / t2**(3./2.)


def _figureOfMerit(t, raw, resids, cc):
    """
    Figure of merit for deconvolution that combines the positivity of the 
    residuals, the skewness of the clean components, the RMS of the 
    residuals, and the number of noise-like points
    
    From:
    Bhat, N., Cordes, J., & Chatterjee, S.  2003, ApJ, 584, 782.
    """

    # Use robust methods to estimate the off-pulse mean and standard deviation
    m = robust.mean(raw)
    s = robust.std(raw)
    
    # Find the fraction of noise-like points in the residuals
    n = len( numpy.where( numpy.abs(resids - m) <= 3*s )[0] )
    n = float(n) / raw.size

    # Compute the positivity and skewness
    f = _positivity(t, raw, resids, cc)
    g = _skewness(t, raw, resids, cc)
    
    # Get the standard deviation of all of the residuals relative to the estimated
    # off-pulse value
    r  = resids.std()
    r /= s
    
    # The figure-of-metric
    return (f + g)/2.0 + r - n


def unscatter(t, raw, tScatMin, tScatMax, tScatStep, gain=0.05, iterMax=10000, broadeningFunction=thin, verbose=True):
    """
    Multi-path scattering deconvolution method based on the method 
    presented in Bhat, N., Cordes, J., & Chatterjee, S.  2003, ApJ, 
    584, 782.
    
    Inputs:
    1) t: List of times in seconds the pulse profile corresponds to
    2) raw: the raw (scattered) pulse profile over time
    3) tScatMin: minimum scattering time to search
    4) tScatMax: maximum scattering time to search
    5) tScatStep: time step for tScat search
        
    Options:
    * gain: CLEAN loop gain (default is 0.05)
    * iterMax: maximum number of iterations to use (default is 10000)
    * broadeningFunction: pulse broadening function (default is thin)
        
    Outputs (as a tuple):
    1) tScat: best-fit scattering time in seconds
    2) merit: figure of merit for the deconvolution/scattering time fit
    3) cleand: CLEANed pulsar profile as a function of time
    """
    
    iList = {}
    meritList = {}
    ccList = {}
    residList = {}

    # Off-pulsar standard deviation
    sigma = robust.std(raw)

    # Loop over tScat values
    best = 1e9
    bestTau = None
    for tScat in numpy.arange(tScatMin, tScatMax, tScatStep):
        ## Setup the temporary variables
        working = raw*1.0
        cc = raw*0.0

        ## Iterate...
        i = 0
        while i < iterMax:
            ### Find the peak and make sure it is really a peak
            peak = numpy.where( working == working.max() )[0][0]
            if working.max() < 3./2.*sigma:
                break

            ### Generate the clean component
            tPeak = t[peak]
            tRel = t - tPeak
            toRemove  = broadeningFunction(tRel, tScat)
            toRemove /= toRemove.sum()
            toRemove *= gain*working.max()

            ### Remove and continue
            cc[peak] += toRemove.sum()
            working -= toRemove
            i += 1
            
        ## Evaluate and save
        iList[tScat] = i
        meritList[tScat] = _figureOfMerit(t, raw, working, cc)
        ccList[tScat] = cc
        residList[tScat] = working
        
        ## Compare
        if meritList[tScat] < best:
            best = meritList[tScat]
            bestTau = tScat
            
    # Make sure we have something to report
    if tScat is None:
        raise RuntimeError("No good solution found, consider changing the search range.")
        
    # Select out what we need, i.e., the best
    tScat = bestTau
    i = iList[bestTau]
    merit = meritList[bestTau]
    cc = ccList[bestTau]
    resids = residList[bestTau]
    
    # Report on the findings
    if verbose:
        print("Multi-path Scattering Results:")
        print("  Iterations Used: %i of %i" % (i, iterMax))
        print("  Best-fit Scattering time: %.3f ms" % (tScat*1000.0,))
        print("  Figure-of-merit:  %.5f" % merit)
        
    # Restore the profile using a Gaussian with a sigma value of 5 time steps
    sigmaRestore = 5*(t[1]-t[0])
    restoreFunction = numpy.exp(-t**2 / (2*sigmaRestore**2))
    restoreFunction /= restoreFunction.sum()
    out = numpy.convolve(cc, restoreFunction, 'same')
    
    # Add back in the residuals
    out += resids
    
    # Done!
    return tScat, merit, out
    